# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# This file is originally part of ranger, the console file manager. https://github.com/ranger/ranger
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import curses
import threading
import time
from functools import partial

from nvitop.gui.library.displayable import Displayable
from nvitop.gui.library.keybinding import normalize_keybinding
from nvitop.gui.library.process import host
from nvitop.gui.library.utils import cut_string
from nvitop.gui.library.widestring import WideString


class MessageBox(Displayable):  # pylint: disable=too-many-instance-attributes
    class Option:  # pylint: disable=too-few-public-methods
        # pylint: disable-next=too-many-arguments
        def __init__(self, name, key, callback, keys=(), attrs=()):
            self.name = WideString(name)
            self.offset = 0
            self.key = normalize_keybinding(key)
            self.callback = callback
            self.keys = tuple({normalize_keybinding(key) for key in keys}.difference({self.key}))
            self.attrs = attrs

        def __str__(self):
            return str(self.name)

    # pylint: disable-next=too-many-arguments
    def __init__(self, message, options, default, yes, no, cancel, win, root):
        super().__init__(win, root)

        if default is None:
            default = 0
        if no is None:
            no = cancel

        self.options = options
        self.num_options = len(self.options)

        assert cancel is not None
        assert self.num_options >= 2
        assert 0 <= no < self.num_options
        assert 0 <= cancel < self.num_options
        assert 0 <= default < self.num_options

        self.previous_focused = None
        self.message = message
        self.previous_keymap = root.keymaps.used_keymap
        self.current = default
        self.yes = yes
        self.cancel = cancel
        self.no = no  # pylint: disable=invalid-name
        self.timestamp = time.monotonic()

        self.name_len = max(8, max(len(option.name) for option in options))
        for option in self.options:
            option.offset = (self.name_len - len(option.name)) // 2
            option.name = option.name.center(self.name_len)

        self.xy_mouse = None
        self.x, self.y = root.x, root.y
        self.width = (self.name_len + 6) * self.num_options + 6

        self.init_keybindings()

        lines = []
        for msg in self.message.splitlines():
            words = iter(map(WideString, msg.split()))
            try:
                lines.append(next(words))
            except StopIteration:
                lines.append('')
                continue
            for word in words:
                if len(lines[-1]) + len(word) + 1 <= self.width - 6:
                    lines[-1] += ' ' + word
                else:
                    lines[-1] = lines[-1].strip()
                    lines.append(word)
        if len(lines) == 1:
            lines[-1] = WideString(lines[-1]).center(self.width - 6)
        lines = [f' │ {line.ljust(self.width - 6)} │ ' for line in lines]
        lines = [
            ' ╒' + '═' * (self.width - 4) + '╕ ',
            ' │' + ' ' * (self.width - 4) + '│ ',
            *lines,
            ' │' + ' ' * (self.width - 4) + '│ ',
            ' │  ' + '  '.join(['┌' + '─' * (self.name_len + 2) + '┐'] * self.num_options) + '  │ ',
            ' │  ' + '  '.join(map('│ {} │'.format, self.options)) + '  │ ',
            ' │  ' + '  '.join(['└' + '─' * (self.name_len + 2) + '┘'] * self.num_options) + '  │ ',
            ' ╘' + '═' * (self.width - 4) + '╛ ',
        ]
        self.lines = lines

    @property
    def current(self):
        return self._current

    @current.setter
    def current(self, value):
        self._current = value
        self.timestamp = time.monotonic()

    def draw(self):
        self.set_base_attr(attr=0)
        self.color_reset()

        n_term_lines, n_term_cols = self.root.termsize

        height = len(self.lines)
        y_start, x_start = (n_term_lines - height) // 2, (n_term_cols - self.width) // 2
        y_option_start = y_start + height - 3
        for y, line in enumerate(self.lines, start=y_start):
            self.addstr(y, x_start, line)

        for i, option in enumerate(self.options):
            x_option_start = x_start + 6 + i * (self.name_len + 6) + option.offset
            for attr in option.attrs:
                attr = attr.copy()
                y = y_option_start + attr.pop('y')
                x = x_option_start + attr.pop('x')
                self.color_at(y, x, **attr)

        if self.xy_mouse is not None:
            x, y = self.xy_mouse
            if y_option_start - 1 <= y <= y_option_start + 1:
                current = (x - x_start - 3) // (self.name_len + 6)
                x_option_start = x_start + 6 + current * (self.name_len + 6)
                if (
                    0 <= current < self.num_options
                    and x_option_start - 3 <= x < x_option_start + self.name_len + 3
                ):
                    self.apply(current, wait=True)

        option = self.options[self.current]
        x_option_start = x_start + 6 + self.current * (self.name_len + 6)
        for y in range(y_option_start - 1, y_option_start + 2):
            self.color_at(
                y,
                x_option_start - 3,
                width=self.name_len + 6,
                attr='standout | bold',
            )
        for attr in option.attrs:
            attr = attr.copy()
            y = y_option_start + attr.pop('y')
            x = x_option_start + option.offset + attr.pop('x')
            attr['fg'], attr['bg'] = attr.get('bg', -1), attr.get('fg', -1)
            attr['attr'] = self.get_fg_bg_attr(attr=attr.get('attr', 0))
            attr['attr'] |= self.get_fg_bg_attr(attr='standout | bold')
            self.color_at(y, x, **attr)

    def finalize(self):
        self.xy_mouse = None
        super().finalize()

    def press(self, key):
        self.root.keymaps.use_keymap('messagebox')
        self.root.press(key)

    def click(self, event):
        if event.pressed(1) or event.pressed(3) or event.clicked(1) or event.clicked(3):
            self.xy_mouse = (event.x, event.y)
            return True

        direction = event.wheel_direction()
        self.current = (self.current + direction) % self.num_options
        return True

    def apply(self, index=None, wait=None):
        if index is None:
            index = self.current

        assert 0 <= index < self.num_options

        if (index != self.current and wait is None) or wait:
            self.current = index

            def confirm():
                time.sleep(0.25)
                curses.ungetch(curses.KEY_ENTER)

            threading.Thread(name='messagebox-confirm', target=confirm, daemon=True).start()
            return

        callback = self.options[index].callback
        if callback is not None:
            callback()

        self.root.keymaps.clear_keymap('messagebox')
        self.root.keymaps.use_keymap(self.previous_keymap)
        self.root.need_redraw = True
        self.root.messagebox = None

    def init_keybindings(self):  # pylint: disable=too-many-branches
        def select_previous():
            self.current = (self.current - 1) % self.num_options

        def select_next():
            self.current = (self.current + 1) % self.num_options

        keymaps = self.root.keymaps
        keymaps.clear_keymap('messagebox')

        for i, option in enumerate(self.options):
            keymaps.bind('messagebox', option.key, partial(self.apply, index=i))
            for key in option.keys:
                keymaps.copy('messagebox', option.key, key)

        keymaps['messagebox'][keymaps.keybuffer.quantifier_key] = 'false'
        if (
            len(set('0123456789').intersection(keymaps['messagebox'])) == 0
            and self.num_options <= 9
        ):
            for key_n, option in zip('123456789', self.options):
                keymaps.copy('messagebox', option.key, key_n)

        assert (
            len({'<Enter>', '<Esc>', '<Left>', '<Right>'}.intersection(keymaps['messagebox'])) == 0
        )

        if self.yes is not None and 'y' not in keymaps['messagebox']:
            keymaps.copy('messagebox', self.options[self.yes].key, 'y')
            if 'Y' not in keymaps['messagebox']:
                keymaps.copy('messagebox', self.options[self.yes].key, 'Y')
        if self.no is not None and 'n' not in keymaps['messagebox']:
            keymaps.copy('messagebox', self.options[self.no].key, 'n')
            if 'N' not in keymaps['messagebox']:
                keymaps.copy('messagebox', self.options[self.no].key, 'N')
        if self.cancel is not None:
            keymaps.bind('messagebox', '<Esc>', partial(self.apply, index=self.cancel, wait=False))
            if 'q' not in keymaps['messagebox'] and 'Q' not in keymaps['messagebox']:
                keymaps.copy('messagebox', '<Esc>', 'q')
                keymaps.copy('messagebox', '<Esc>', 'Q')

        keymaps.bind('messagebox', '<Enter>', self.apply)
        if '<Space>' not in keymaps['messagebox']:
            keymaps.copy('messagebox', '<Enter>', '<Space>')

        keymaps.bind('messagebox', '<Left>', select_previous)
        keymaps.bind('messagebox', '<Right>', select_next)
        if ',' not in keymaps['messagebox'] and '.' not in keymaps['messagebox']:
            keymaps.copy('messagebox', '<Left>', ',')
            keymaps.copy('messagebox', '<Right>', '.')
        if '<' not in keymaps['messagebox'] and '>' not in keymaps['messagebox']:
            keymaps.copy('messagebox', '<Left>', '<')
            keymaps.copy('messagebox', '<Right>', '>')
        if '[' not in keymaps['messagebox'] and ']' not in keymaps['messagebox']:
            keymaps.copy('messagebox', '<Left>', '[')
            keymaps.copy('messagebox', '<Right>', ']')
        if '<Tab>' not in keymaps['messagebox'] and '<S-Tab>' not in keymaps['messagebox']:
            keymaps.copy('messagebox', '<Left>', '<S-Tab>')
            keymaps.copy('messagebox', '<Right>', '<Tab>')


def send_signal(signal, panel):
    assert signal in ('terminate', 'kill', 'interrupt')
    default = {'terminate': 0, 'kill': 1, 'interrupt': 2}.get(signal)
    processes = []
    for process in panel.selection.processes():
        try:
            username = process.username()
        except host.PsutilError:
            username = 'N/A'
        username = cut_string(username, maxlen=24, padstr='+')
        processes.append(f'{process.pid}({username})')
    if len(processes) == 0:
        return
    if len(processes) == 1:
        message = f'Send signal to process {processes[0]}?'
    else:
        maxlen = max(map(len, processes))
        processes = [process.ljust(maxlen) for process in processes]
        message = 'Send signal to the following processes?\n\n{}'.format(' '.join(processes))

    # pylint: disable=use-dict-literal
    panel.root.messagebox = MessageBox(
        message=message,
        options=[
            MessageBox.Option(
                'SIGTERM',
                't',
                panel.selection.terminate,
                keys=('T',),
                attrs=(
                    {'y': 0, 'x': 0, 'width': 7, 'fg': 'red'},
                    {'y': 0, 'x': 3, 'width': 1, 'fg': 'red', 'attr': 'bold | underline'},
                ),
            ),
            MessageBox.Option(
                'SIGKILL',
                'k',
                panel.selection.kill,
                keys=('K',),
                attrs=(
                    {'y': 0, 'x': 0, 'width': 7, 'fg': 'red'},
                    {'y': 0, 'x': 3, 'width': 1, 'fg': 'red', 'attr': 'bold | underline'},
                ),
            ),
            MessageBox.Option(
                'SIGINT',
                'i',
                panel.selection.interrupt,
                keys=('I',),
                attrs=(
                    {'y': 0, 'x': 0, 'width': 6, 'fg': 'red'},
                    {'y': 0, 'x': 3, 'width': 1, 'fg': 'red', 'attr': 'bold | underline'},
                ),
            ),
            MessageBox.Option(
                'Cancel',
                'c',
                None,
                keys=('C',),
                attrs=({'y': 0, 'x': 0, 'width': 1, 'attr': 'bold | underline'},),
            ),
        ],
        default=default,
        yes=None,
        no=3,
        cancel=3,
        win=panel.win,
        root=panel.root,
    )
    # pylint: enable=use-dict-literal



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import curses
import shutil
import time

from nvitop.gui.library import ALT_KEY, DisplayableContainer, KeyBuffer, KeyMaps, MouseEvent
from nvitop.gui.screens import (
    BreakLoop,
    EnvironScreen,
    HelpScreen,
    MainScreen,
    ProcessMetricsScreen,
    TreeViewScreen,
)


class UI(DisplayableContainer):  # pylint: disable=too-many-instance-attributes
    # pylint: disable-next=too-many-arguments
    def __init__(
        self,
        devices,
        filters=(),
        ascii=False,  # pylint: disable=redefined-builtin
        mode='auto',
        interval=None,
        win=None,
    ):
        super().__init__(win, root=self)

        self.x = self.y = 0
        self.width = max(79, shutil.get_terminal_size(fallback=(79, 24)).columns - self.x)
        self.termsize = None

        self.ascii = ascii

        self.devices = devices
        self.device_count = len(self.devices)

        self.main_screen = MainScreen(
            self.devices,
            filters,
            ascii=ascii,
            mode=mode,
            win=win,
            root=self,
        )
        self.main_screen.visible = True
        self.main_screen.focused = False
        self.add_child(self.main_screen)
        self.current_screen = self.previous_screen = self.main_screen

        self._messagebox = None

        if win is not None:
            self.environ_screen = EnvironScreen(win=win, root=self)
            self.environ_screen.visible = False
            self.environ_screen.ascii = False
            self.add_child(self.environ_screen)

            self.treeview_screen = TreeViewScreen(win=win, root=self)
            self.treeview_screen.visible = False
            self.treeview_screen.ascii = self.ascii
            self.add_child(self.treeview_screen)

            self.process_metrics_screen = ProcessMetricsScreen(win=win, root=self)
            self.process_metrics_screen.visible = False
            self.process_metrics_screen.ascii = self.ascii
            self.add_child(self.process_metrics_screen)

            self.help_screen = HelpScreen(win=win, root=self)
            self.help_screen.visible = False
            self.help_screen.ascii = False
            self.add_child(self.help_screen)

            if interval is not None:
                if interval < 1.0:
                    self.main_screen.device_panel.set_snapshot_interval(interval)
                    self.main_screen.host_panel.set_snapshot_interval(interval)
                if interval < 0.5:
                    self.process_metrics_screen.set_snapshot_interval(interval)
                self.main_screen.process_panel.set_snapshot_interval(interval)
                self.treeview_screen.set_snapshot_interval(interval)

            self.keybuffer = KeyBuffer()
            self.keymaps = KeyMaps(self.keybuffer)
            self.last_input_time = time.monotonic()
            self.init_keybindings()

    @property
    def messagebox(self):
        return self._messagebox

    @messagebox.setter
    def messagebox(self, value):
        self.need_redraw = True
        if self._messagebox is not None:
            self.remove_child(self._messagebox)

        self._messagebox = value
        if value is not None:
            self._messagebox.visible = True
            self._messagebox.focused = True
            self._messagebox.ascii = self.ascii
            self._messagebox.previous_focused = self.get_focused_obj()
            self.add_child(self._messagebox)

    def get_focused_obj(self):
        if self.messagebox is not None:
            return self.messagebox
        return super().get_focused_obj()

    def update_size(self, termsize=None):
        n_term_lines, n_term_cols = termsize = super().update_size(termsize=termsize)

        self.width = n_term_cols - self.x
        self.height = n_term_lines - self.y

        for screen in self.container:
            if hasattr(screen, 'update_size'):
                screen.update_size(termsize)

        if self.termsize != termsize:
            self.termsize = termsize
            self.need_redraw = True

        return termsize

    def poke(self):
        super().poke()

        if self.termsize is None:
            self.update_size()

    def draw(self):
        if self.need_redraw:
            self.win.erase()

        self.set_base_attr(attr=0)
        self.color_reset()

        if self.width >= 79:
            if self.messagebox is not None:
                self.set_base_attr(attr='dim')
            super().draw()
            return
        if not self.need_redraw:
            return

        n_term_lines, n_term_cols = self.termsize
        message = (
            f'nvitop needs at least a width of 79 to render, the current width is {self.width}.'
        )
        words = iter(message.split())
        width = min(max(n_term_cols, 40), n_term_cols, 60) - 10
        lines = [next(words)]
        for word in words:
            if len(lines[-1]) + len(word) + 1 <= width:
                lines[-1] += ' ' + word
            else:
                lines[-1] = lines[-1].strip()
                lines.append(word)
        height, width = len(lines) + 4, max(map(len, lines)) + 4
        lines = [f'│ {line.ljust(width - 4)} │' for line in lines]
        lines = [
            '╒' + '═' * (width - 2) + '╕',
            '│' + ' ' * (width - 2) + '│',
            *lines,
            '│' + ' ' * (width - 2) + '│',
            '╘' + '═' * (width - 2) + '╛',
        ]

        y_start, x_start = (n_term_lines - height) // 2, (n_term_cols - width) // 2
        for y, line in enumerate(lines, start=y_start):
            self.addstr(y, x_start, line)

    def finalize(self):
        super().finalize()
        self.win.refresh()

    def redraw(self):
        self.poke()
        self.draw()
        self.finalize()

    def loop(self):
        if self.win is None:
            return

        while True:
            try:
                self.redraw()
                self.handle_input()
                if time.monotonic() - self.last_input_time > 1.0:
                    time.sleep(0.2)
            except BreakLoop:
                break

    def print(self):
        self.main_screen.print()

    def handle_mouse(self):
        """Handle mouse input."""
        try:
            event = MouseEvent(curses.getmouse())
        except curses.error:
            return
        super().click(event)

    def handle_key(self, key):
        """Handle key input."""
        if key < 0:
            self.keybuffer.clear()
        elif not super().press(key):
            self.keymaps.use_keymap('main')
            self.press(key)

    def handle_keys(self, *keys):
        for key in keys:
            self.handle_key(key)

    def press(self, key):
        keybuffer = self.keybuffer

        keybuffer.add(key)
        if keybuffer.result is not None:
            try:
                keybuffer.result()
            finally:
                if keybuffer.finished_parsing:
                    keybuffer.clear()
        elif keybuffer.finished_parsing:
            keybuffer.clear()
            return False
        return True

    def handle_input(self):  # pylint: disable=too-many-branches
        key = self.win.getch()
        if key == curses.ERR:
            return

        self.last_input_time = time.monotonic()
        if key == curses.KEY_ENTER:
            key = ord('\n')
        if key == 27 or (128 <= key < 256):
            # Handle special keys like ALT+X or unicode here:
            keys = [key]
            for _ in range(4):
                getkey = self.win.getch()
                if getkey != -1:
                    keys.append(getkey)
            if len(keys) == 1:
                keys.append(-1)
            elif keys[0] == 27:
                keys[0] = ALT_KEY
            self.handle_keys(*keys)
            curses.flushinp()
        elif key >= 0:
            # Handle simple key presses, CTRL+X, etc here:
            curses.flushinp()
            if key == curses.KEY_MOUSE:
                self.handle_mouse()
            elif key == curses.KEY_RESIZE:
                self.update_size()
            else:
                self.handle_key(key)

    def init_keybindings(self):
        # pylint: disable=multiple-statements,too-many-statements

        for screen in self.container:
            if hasattr(screen, 'init_keybindings'):
                screen.init_keybindings()

        def show_screen(screen, focused=None):
            for s in self.container:
                if s is screen:
                    s.visible = True
                    if focused is not None:
                        s.focused = focused
                else:
                    s.visible = False

            self.previous_screen = self.current_screen
            self.current_screen = screen

        def show_main():
            show_screen(self.main_screen, focused=False)

            if self.treeview_screen.selection.is_set():
                self.main_screen.selection.process = self.treeview_screen.selection.process
            self.treeview_screen.selection.clear()
            self.process_metrics_screen.disable()

        def show_environ():
            show_screen(self.environ_screen, focused=True)

            if self.previous_screen is not self.help_screen:
                self.environ_screen.process = self.previous_screen.selection.process

        def environ_return():
            if self.previous_screen is self.treeview_screen:
                show_treeview()
            elif self.previous_screen is self.process_metrics_screen:
                show_process_metrics()
            else:
                show_main()

        def show_treeview():
            if not self.main_screen.process_panel.has_snapshots:
                return

            show_screen(self.treeview_screen, focused=True)

            if not self.treeview_screen.selection.is_set():
                self.treeview_screen.selection.process = self.main_screen.selection.process
            self.main_screen.selection.clear()

        def show_process_metrics():
            if self.current_screen is self.main_screen:
                if self.main_screen.selection.is_set():
                    show_screen(self.process_metrics_screen, focused=True)
                    self.process_metrics_screen.process = self.previous_screen.selection.process
            elif self.current_screen is not self.treeview_screen:
                show_screen(self.process_metrics_screen, focused=True)

        def show_help():
            show_screen(self.help_screen, focused=True)

        def help_return():
            if self.previous_screen is self.treeview_screen:
                show_treeview()
            elif self.previous_screen is self.environ_screen:
                show_environ()
            elif self.previous_screen is self.process_metrics_screen:
                show_process_metrics()
            else:
                show_main()

        self.keymaps.bind('main', 'e', show_environ)
        self.keymaps.bind('environ', 'e', environ_return)
        self.keymaps.copy('environ', 'e', '<Esc>')
        self.keymaps.copy('environ', 'e', 'q')
        self.keymaps.copy('environ', 'e', 'Q')

        self.keymaps.bind('main', 't', show_treeview)
        self.keymaps.bind('treeview', 't', show_main)
        self.keymaps.copy('treeview', 't', 'q')
        self.keymaps.copy('treeview', 't', 'Q')
        self.keymaps.bind('treeview', 'e', show_environ)

        self.keymaps.bind('main', '<Enter>', show_process_metrics)
        self.keymaps.bind('process-metrics', '<Enter>', show_main)
        self.keymaps.copy('process-metrics', '<Enter>', '<Esc>')
        self.keymaps.copy('process-metrics', '<Enter>', 'q')
        self.keymaps.copy('process-metrics', '<Enter>', 'Q')
        self.keymaps.bind('process-metrics', 'e', show_environ)

        for screen in ('main', 'treeview', 'environ', 'process-metrics'):
            self.keymaps.bind(screen, 'h', show_help)
            self.keymaps.copy(screen, 'h', '?')
        self.keymaps.bind('help', '<Esc>', help_return)
        self.keymaps.bind('help', '<any>', help_return)

        self.keymaps.use_keymap('main')



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import threading
import time

from nvitop.gui.library import (
    NA,
    BufferedHistoryGraph,
    Device,
    Displayable,
    GiB,
    bytes2human,
    colored,
    host,
    make_bar,
    timedelta2human,
)


class HostPanel(Displayable):  # pylint: disable=too-many-instance-attributes
    NAME = 'host'
    SNAPSHOT_INTERVAL = 0.5

    def __init__(self, devices, compact, win, root):
        super().__init__(win, root)

        self.devices = devices
        self.device_count = len(self.devices)

        if win is not None:
            self.average_gpu_memory_percent = None
            self.average_gpu_utilization = None
            self.enable_history()

        self._compact = compact
        self.width = max(79, root.width)
        self.full_height = 12
        self.compact_height = 2
        self.height = self.compact_height if compact else self.full_height

        self.cpu_percent = None
        self.load_average = None
        self.virtual_memory = None
        self.swap_memory = None
        self._snapshot_daemon = threading.Thread(
            name='host-snapshot-daemon',
            target=self._snapshot_target,
            daemon=True,
        )
        self._daemon_running = threading.Event()

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        width = max(79, value)
        if self._width != width:
            if self.visible:
                self.need_redraw = True
            graph_width = max(width - 80, 20)
            if self.win is not None:
                self.average_gpu_memory_percent.width = graph_width
                self.average_gpu_utilization.width = graph_width
                for device in self.devices:
                    device.memory_percent.history.width = graph_width
                    device.gpu_utilization.history.width = graph_width
        self._width = width

    @property
    def compact(self):
        return self._compact or self.ascii

    @compact.setter
    def compact(self, value):
        value = value or self.ascii
        if self._compact != value:
            self.need_redraw = True
            self._compact = value
            self.height = self.compact_height if self.compact else self.full_height

    def enable_history(self):
        host.cpu_percent = BufferedHistoryGraph(
            interval=1.0,
            width=77,
            height=5,
            upsidedown=False,
            baseline=0.0,
            upperbound=100.0,
            dynamic_bound=False,
            format='CPU: {:.1f}%'.format,
        )(host.cpu_percent)
        host.virtual_memory = BufferedHistoryGraph(
            interval=1.0,
            width=77,
            height=4,
            upsidedown=True,
            baseline=0.0,
            upperbound=100.0,
            dynamic_bound=False,
            format='{:.1f}%'.format,
        )(host.virtual_memory, get_value=lambda vm: vm.percent)
        host.swap_memory = BufferedHistoryGraph(
            interval=1.0,
            width=77,
            height=1,
            upsidedown=False,
            baseline=0.0,
            upperbound=100.0,
            dynamic_bound=False,
            format='{:.1f}%'.format,
        )(host.swap_memory, get_value=lambda sm: sm.percent)

        def percentage(x):
            return f'{x:.1f}%' if x is not NA else NA

        def enable_history(device):
            device.memory_percent = BufferedHistoryGraph(
                interval=1.0,
                width=20,
                height=5,
                upsidedown=False,
                baseline=0.0,
                upperbound=100.0,
                dynamic_bound=False,
                format=lambda x: f'GPU {device.display_index} MEM: {percentage(x)}',
            )(device.memory_percent)
            device.gpu_utilization = BufferedHistoryGraph(
                interval=1.0,
                width=20,
                height=5,
                upsidedown=True,
                baseline=0.0,
                upperbound=100.0,
                dynamic_bound=False,
                format=lambda x: f'GPU {device.display_index} UTL: {percentage(x)}',
            )(device.gpu_utilization)

        for device in self.devices:
            enable_history(device)

        prefix = 'AVG ' if self.device_count > 1 else ''
        self.average_gpu_memory_percent = BufferedHistoryGraph(
            interval=1.0,
            width=20,
            height=5,
            upsidedown=False,
            baseline=0.0,
            upperbound=100.0,
            dynamic_bound=False,
            format=lambda x: f'{prefix}GPU MEM: {percentage(x)}',
        )
        self.average_gpu_utilization = BufferedHistoryGraph(
            interval=1.0,
            width=20,
            height=5,
            upsidedown=True,
            baseline=0.0,
            upperbound=100.0,
            dynamic_bound=False,
            format=lambda x: f'{prefix}GPU UTL: {percentage(x)}',
        )

    @classmethod
    def set_snapshot_interval(cls, interval):
        assert interval > 0.0
        interval = float(interval)

        cls.SNAPSHOT_INTERVAL = min(interval / 3.0, 0.5)

    def take_snapshots(self):
        host.cpu_percent()
        host.virtual_memory()
        host.swap_memory()
        self.load_average = host.load_average()

        self.cpu_percent = host.cpu_percent.history.last_value
        self.virtual_memory = host.virtual_memory.history.last_retval
        self.swap_memory = host.swap_memory.history.last_retval

        total_memory_used = 0
        total_memory_total = 0
        gpu_utilizations = []
        for device in self.devices:
            memory_used = device.snapshot.memory_used
            memory_total = device.snapshot.memory_total
            gpu_utilization = device.snapshot.gpu_utilization
            if memory_used is not NA and memory_total is not NA:
                total_memory_used += memory_used
                total_memory_total += memory_total
            if gpu_utilization is not NA:
                gpu_utilizations.append(float(gpu_utilization))
        if total_memory_total > 0:
            self.average_gpu_memory_percent.add(100.0 * total_memory_used / total_memory_total)
        if len(gpu_utilizations) > 0:
            self.average_gpu_utilization.add(sum(gpu_utilizations) / len(gpu_utilizations))

    def _snapshot_target(self):
        self._daemon_running.wait()
        while self._daemon_running.is_set():
            self.take_snapshots()
            time.sleep(self.SNAPSHOT_INTERVAL)

    def frame_lines(self, compact=None):
        if compact is None:
            compact = self.compact
        if compact or self.ascii:
            return []

        remaining_width = self.width - 79
        data_line = (
            '│                                                                             │'
        )
        separator_line = (
            '├────────────╴120s├─────────────────────────╴60s├──────────╴30s├──────────────┤'
        )
        if self.width >= 100:
            data_line += ' ' * (remaining_width - 1) + '│'
            separator_line = separator_line[:-1] + '┼' + '─' * (remaining_width - 1) + '┤'

        frame = [
            '╞═══════════════════════════════╧══════════════════════╧══════════════════════╡',
            data_line,
            data_line,
            data_line,
            data_line,
            data_line,
            separator_line,
            data_line,
            data_line,
            data_line,
            data_line,
            data_line,
            '╘═════════════════════════════════════════════════════════════════════════════╛',
        ]
        if self.width >= 100:
            frame[0] = frame[0][:-1] + '╪' + '═' * (remaining_width - 1) + '╡'
            frame[-1] = frame[-1][:-1] + '╧' + '═' * (remaining_width - 1) + '╛'

        return frame

    def poke(self):
        if not self._daemon_running.is_set():
            self._daemon_running.set()
            self._snapshot_daemon.start()
            self.take_snapshots()

        super().poke()

    def draw(self):  # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        self.color_reset()

        if self.load_average is not None:
            load_average = tuple(
                f'{value:5.2f}'[:5] if value < 10000.0 else '9999+' for value in self.load_average
            )
        else:
            load_average = (NA,) * 3
        load_average = 'Load Average: {} {} {}'.format(*load_average)

        if self.compact:
            width_right = len(load_average) + 4
            width_left = self.width - 2 - width_right
            cpu_bar = '[ {} ]'.format(
                make_bar(
                    'CPU',
                    self.cpu_percent,
                    width_left - 4,
                    extra_text=f'  UPTIME: {timedelta2human(host.uptime(), round=True)}',
                ),
            )
            memory_bar = '[ {} ]'.format(
                make_bar(
                    'MEM',
                    self.virtual_memory.percent,
                    width_left - 4,
                    extra_text=f'  USED: {bytes2human(self.virtual_memory.used, min_unit=GiB)}',
                ),
            )
            swap_bar = '[ {} ]'.format(make_bar('SWP', self.swap_memory.percent, width_right - 4))
            self.addstr(self.y, self.x, f'{cpu_bar}  ( {load_average} )')
            self.addstr(self.y + 1, self.x, f'{memory_bar}  {swap_bar}')
            self.color_at(self.y, self.x, width=len(cpu_bar), fg='cyan', attr='bold')
            self.color_at(self.y + 1, self.x, width=width_left, fg='magenta', attr='bold')
            self.color_at(self.y, self.x + width_left + 2, width=width_right, attr='bold')
            self.color_at(
                self.y + 1,
                self.x + width_left + 2,
                width=width_right,
                fg='blue',
                attr='bold',
            )
            return

        remaining_width = self.width - 79

        if self.need_redraw:
            for y, line in enumerate(self.frame_lines(), start=self.y - 1):
                self.addstr(y, self.x, line)
            self.color_at(self.y + 5, self.x + 14, width=4, attr='dim')
            self.color_at(self.y + 5, self.x + 45, width=3, attr='dim')
            self.color_at(self.y + 5, self.x + 60, width=3, attr='dim')

            if self.width >= 100:
                for offset, string in (
                    (20, '╴30s├'),
                    (35, '╴60s├'),
                    (66, '╴120s├'),
                    (96, '╴180s├'),
                    (126, '╴240s├'),
                    (156, '╴300s├'),
                ):
                    if offset > remaining_width:
                        break
                    self.addstr(self.y + 5, self.x + self.width - offset, string)
                    self.color_at(
                        self.y + 5,
                        self.x + self.width - offset + 1,
                        width=len(string) - 2,
                        attr='dim',
                    )

        self.color(fg='cyan')
        for y, line in enumerate(host.cpu_percent.history.graph, start=self.y):
            self.addstr(y, self.x + 1, line)

        self.color(fg='magenta')
        for y, line in enumerate(host.virtual_memory.history.graph, start=self.y + 6):
            self.addstr(y, self.x + 1, line)

        self.color(fg='blue')
        for y, line in enumerate(host.swap_memory.history.graph, start=self.y + 10):
            self.addstr(y, self.x + 1, line)

        if self.width >= 100:
            if self.device_count > 1 and self.parent.selection.is_set():
                device = self.parent.selection.process.device
                gpu_memory_percent = device.memory_percent.history
                gpu_utilization = device.gpu_utilization.history
            else:
                gpu_memory_percent = self.average_gpu_memory_percent
                gpu_utilization = self.average_gpu_utilization

            if self.TERM_256COLOR:
                for i, (y, line) in enumerate(enumerate(gpu_memory_percent.graph, start=self.y)):
                    self.addstr(y, self.x + 79, line, self.get_fg_bg_attr(fg=1.0 - i / 4.0))

                for i, (y, line) in enumerate(enumerate(gpu_utilization.graph, start=self.y + 6)):
                    self.addstr(y, self.x + 79, line, self.get_fg_bg_attr(fg=i / 4.0))
            else:
                self.color(fg=Device.color_of(gpu_memory_percent.last_value, type='memory'))
                for y, line in enumerate(gpu_memory_percent.graph, start=self.y):
                    self.addstr(y, self.x + 79, line)

                self.color(fg=Device.color_of(gpu_utilization.last_value, type='gpu'))
                for y, line in enumerate(gpu_utilization.graph, start=self.y + 6):
                    self.addstr(y, self.x + 79, line)

        self.color_reset()
        self.addstr(self.y, self.x + 1, f' {load_average} ')
        self.addstr(self.y + 1, self.x + 1, f' {host.cpu_percent.history} ')
        self.addstr(
            self.y + 9,
            self.x + 1,
            f' MEM: {bytes2human(self.virtual_memory.used, min_unit=GiB)} ({host.virtual_memory.history}) ',
        )
        self.addstr(
            self.y + 10,
            self.x + 1,
            f' SWP: {bytes2human(self.swap_memory.used, min_unit=GiB)} ({host.swap_memory.history}) ',
        )
        if self.width >= 100:
            self.addstr(self.y, self.x + 79, f' {gpu_memory_percent} ')
            self.addstr(self.y + 10, self.x + 79, f' {gpu_utilization} ')

    def destroy(self):
        super().destroy()
        self._daemon_running.clear()

    def print_width(self):
        if self.device_count > 0 and self.width >= 100:
            return self.width
        return 79

    def print(self):
        self.cpu_percent = host.cpu_percent()
        self.virtual_memory = host.virtual_memory()
        self.swap_memory = host.swap_memory()
        self.load_average = host.load_average()

        if self.load_average is not None:
            load_average = tuple(
                f'{value:5.2f}'[:5] if value < 10000.0 else '9999+' for value in self.load_average
            )
        else:
            load_average = (NA,) * 3
        load_average = 'Load Average: {} {} {}'.format(*load_average)

        width_right = len(load_average) + 4
        width_left = self.width - 2 - width_right
        cpu_bar = '[ {} ]'.format(
            make_bar(
                'CPU',
                self.cpu_percent,
                width_left - 4,
                extra_text=f'  UPTIME: {timedelta2human(host.uptime(), round=True)}',
            ),
        )
        memory_bar = '[ {} ]'.format(
            make_bar(
                'MEM',
                self.virtual_memory.percent,
                width_left - 4,
                extra_text=f'  USED: {bytes2human(self.virtual_memory.used, min_unit=GiB)}',
            ),
        )
        swap_bar = '[ {} ]'.format(make_bar('SWP', self.swap_memory.percent, width_right - 4))

        lines = [
            '{}  {}'.format(
                colored(cpu_bar, color='cyan', attrs=('bold',)),
                colored(f'( {load_average} )', attrs=('bold',)),
            ),
            '{}  {}'.format(
                colored(memory_bar, color='magenta', attrs=('bold',)),
                colored(swap_bar, color='blue', attrs=('bold',)),
            ),
        ]

        lines = '\n'.join(lines)
        if self.ascii:
            lines = lines.translate(self.ASCII_TRANSTABLE)

        try:
            print(lines)
        except UnicodeError:
            print(lines.translate(self.ASCII_TRANSTABLE))

    def press(self, key):
        self.root.keymaps.use_keymap('host')
        self.root.press(key)



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
#
# Copyright 2021-2023 Xuehai Pan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for the NVML Python bindings (`nvidia-ml-py <https://pypi.org/project/nvidia-ml-py>`_)."""

# pylint: disable=invalid-name

from __future__ import annotations

import atexit as _atexit
import ctypes as _ctypes
import inspect as _inspect
import logging as _logging
import os as _os
import re as _re
import sys as _sys
import threading as _threading
import time as _time
from types import FunctionType as _FunctionType
from types import ModuleType as _ModuleType
from typing import TYPE_CHECKING as _TYPE_CHECKING
from typing import Any as _Any
from typing import Callable as _Callable
from typing import ClassVar as _ClassVar

# Python Bindings for the NVIDIA Management Library (NVML)
# https://pypi.org/project/nvidia-ml-py
import pynvml as _pynvml
from pynvml import *  # noqa: F403 # pylint: disable=wildcard-import,unused-wildcard-import
from pynvml import nvmlDeviceGetPciInfo  # appease mypy # noqa: F401 # pylint: disable=unused-import

from nvitop.api.utils import NA, UINT_MAX, ULONGLONG_MAX, NaType
from nvitop.api.utils import colored as __colored


if _TYPE_CHECKING:
    from typing_extensions import Self as _Self  # Python 3.11+
    from typing_extensions import TypeAlias as _TypeAlias  # Python 3.10+


__all__ = [  # will be updated in below
    'NA',
    'UINT_MAX',
    'ULONGLONG_MAX',
    'nvmlCheckReturn',
    'nvmlQuery',
    'nvmlQueryFieldValues',
    'nvmlInit',
    'nvmlInitWithFlags',
    'nvmlShutdown',
    'NVMLError',
]


if not callable(getattr(_pynvml, 'nvmlInitWithFlags', None)):
    raise ImportError(
        'Your installed package `nvidia-ml-py` is corrupted. Please reinstall package '
        '`nvidia-ml-py` via `pip3 install --force-reinstall nvidia-ml-py nvitop`.',
    )


# Members from `pynvml` ############################################################################

NVMLError: type[_pynvml.NVMLError] = _pynvml.NVMLError
NVMLError.__doc__ = """Base exception class for NVML query errors."""
NVMLError.__new__.__doc__ = """Map value to a proper subclass of :class:`NVMLError`."""
nvmlExceptionClass: _Callable[[int], type[_pynvml.NVMLError]] = _pynvml.nvmlExceptionClass
nvmlExceptionClass.__doc__ = """Map value to a proper subclass of :class:`NVMLError`."""

# Load members from module `pynvml` and register them in `__all__` and globals.
_vars_pynvml = vars(_pynvml)
_name = _attr = None
_errcode_to_name = {}
_const_names = []
_errcode_to_string = NVMLError._errcode_to_string  # pylint: disable=protected-access

# 1. Put error classes in `__all__` first
for _name, _attr in _vars_pynvml.items():
    if _name in {'nvmlInit', 'nvmlInitWithFlags', 'nvmlShutdown'}:
        continue
    if _name.startswith(('NVML_ERROR_', 'NVMLError_')):
        __all__.append(_name)  # noqa: PYI056
        if _name.startswith('NVML_ERROR_'):
            _errcode_to_name[_attr] = _name
            _const_names.append(_name)

# 2. Then the remaining members
for _name, _attr in _vars_pynvml.items():
    if _name in {'nvmlInit', 'nvmlInitWithFlags', 'nvmlShutdown'}:
        continue
    if (_name.startswith('NVML_') and not _name.startswith('NVML_ERROR_')) or (
        _name.startswith('nvml') and isinstance(_attr, _FunctionType)
    ):
        __all__.append(_name)  # noqa: PYI056
        if _name.startswith('NVML_'):
            _const_names.append(_name)

# 3. Add docstring to exception classes
_errcode = _reason = _subclass = None
for _errcode, _reason in _errcode_to_string.items():
    _subclass = nvmlExceptionClass(_errcode)
    _subclass.__doc__ = '{}. Code: :data:`{}` ({})'.format(
        _reason.rstrip('.'),
        _errcode_to_name[_errcode],
        _errcode,
    )

# 4. Add undocumented constants into module docstring
_data_docs = []
_sphinx_doc = None
for _name in _const_names:
    _attr = _vars_pynvml[_name]
    _sphinx_doc = f"""
.. data:: {_name}
    :type: {_attr.__class__.__name__}
    :value: {_attr!r}
"""
    if _name.startswith('NVML_ERROR_') and _attr in _errcode_to_string:
        _reason = _errcode_to_string[_attr]
        _sphinx_doc += """
    {}. See also class :class:`NVMLError` and :class:`{}`.
""".format(_reason.rstrip('.'), nvmlExceptionClass(_attr).__name__)  # fmt: skip
    _data_docs.append(_sphinx_doc.strip())
__doc__ += """

---------

Constants
^^^^^^^^^

{}

---------

Functions and Exceptions
^^^^^^^^^^^^^^^^^^^^^^^^

.. function:: __enter__() -> libnvml

    Entry of the context manager for ``with`` statement.

.. function:: __exit__(*args, **kwargs) -> None

    Shutdown the NVML context in the context manager for ``with`` statement.

""".format('\n\n'.join(_data_docs))  # fmt: skip

del (
    _name,
    _attr,
    _vars_pynvml,
    _errcode,
    _reason,
    _subclass,
    _errcode_to_name,
    _errcode_to_string,
    _const_names,
    _data_docs,
    _sphinx_doc,
)

# 5. Add explicit references to appease linters
# pylint: disable=no-member
c_nvmlDevice_t: _TypeAlias = _pynvml.c_nvmlDevice_t  # noqa: PYI042
c_nvmlFieldValue_t: _TypeAlias = _pynvml.c_nvmlFieldValue_t  # noqa: PYI042
NVML_SUCCESS: int = _pynvml.NVML_SUCCESS
NVML_ERROR_INSUFFICIENT_SIZE: int = _pynvml.NVML_ERROR_INSUFFICIENT_SIZE
NVMLError_FunctionNotFound: _TypeAlias = _pynvml.NVMLError_FunctionNotFound
NVMLError_GpuIsLost: _TypeAlias = _pynvml.NVMLError_GpuIsLost
NVMLError_InvalidArgument: _TypeAlias = _pynvml.NVMLError_InvalidArgument
NVMLError_LibraryNotFound: _TypeAlias = _pynvml.NVMLError_LibraryNotFound
NVMLError_NoPermission: _TypeAlias = _pynvml.NVMLError_NoPermission
NVMLError_NotFound: _TypeAlias = _pynvml.NVMLError_NotFound
NVMLError_NotSupported: _TypeAlias = _pynvml.NVMLError_NotSupported
NVMLError_Unknown: _TypeAlias = _pynvml.NVMLError_Unknown
NVML_CLOCK_GRAPHICS: int = _pynvml.NVML_CLOCK_GRAPHICS
NVML_CLOCK_SM: int = _pynvml.NVML_CLOCK_SM
NVML_CLOCK_MEM: int = _pynvml.NVML_CLOCK_MEM
NVML_CLOCK_VIDEO: int = _pynvml.NVML_CLOCK_VIDEO
NVML_TEMPERATURE_GPU: int = _pynvml.NVML_TEMPERATURE_GPU
NVML_DRIVER_WDDM: int = _pynvml.NVML_DRIVER_WDDM
NVML_DRIVER_WDM: int = _pynvml.NVML_DRIVER_WDM
NVML_MEMORY_ERROR_TYPE_UNCORRECTED: int = _pynvml.NVML_MEMORY_ERROR_TYPE_UNCORRECTED
NVML_VOLATILE_ECC: int = _pynvml.NVML_VOLATILE_ECC
NVML_COMPUTEMODE_DEFAULT: int = _pynvml.NVML_COMPUTEMODE_DEFAULT
NVML_COMPUTEMODE_EXCLUSIVE_THREAD: int = _pynvml.NVML_COMPUTEMODE_EXCLUSIVE_THREAD
NVML_COMPUTEMODE_PROHIBITED: int = _pynvml.NVML_COMPUTEMODE_PROHIBITED
NVML_COMPUTEMODE_EXCLUSIVE_PROCESS: int = _pynvml.NVML_COMPUTEMODE_EXCLUSIVE_PROCESS
NVML_PCIE_UTIL_TX_BYTES: int = _pynvml.NVML_PCIE_UTIL_TX_BYTES
NVML_PCIE_UTIL_RX_BYTES: int = _pynvml.NVML_PCIE_UTIL_RX_BYTES
NVML_NVLINK_MAX_LINKS: int = _pynvml.NVML_NVLINK_MAX_LINKS
NVML_FI_DEV_NVLINK_LINK_COUNT: int = _pynvml.NVML_FI_DEV_NVLINK_LINK_COUNT
NVML_FI_DEV_NVLINK_THROUGHPUT_DATA_TX: int = _pynvml.NVML_FI_DEV_NVLINK_THROUGHPUT_DATA_TX
NVML_FI_DEV_NVLINK_THROUGHPUT_DATA_RX: int = _pynvml.NVML_FI_DEV_NVLINK_THROUGHPUT_DATA_RX
NVML_FI_DEV_NVLINK_THROUGHPUT_RAW_TX: int = _pynvml.NVML_FI_DEV_NVLINK_THROUGHPUT_RAW_TX
NVML_FI_DEV_NVLINK_THROUGHPUT_RAW_RX: int = _pynvml.NVML_FI_DEV_NVLINK_THROUGHPUT_RAW_RX
NVML_VALUE_TYPE_DOUBLE: int = getattr(_pynvml, 'NVML_VALUE_TYPE_DOUBLE', 0)
NVML_VALUE_TYPE_UNSIGNED_INT: int = getattr(_pynvml, 'NVML_VALUE_TYPE_UNSIGNED_INT', 1)
NVML_VALUE_TYPE_UNSIGNED_LONG: int = getattr(_pynvml, 'NVML_VALUE_TYPE_UNSIGNED_LONG', 2)
NVML_VALUE_TYPE_UNSIGNED_LONG_LONG: int = getattr(_pynvml, 'NVML_VALUE_TYPE_UNSIGNED_LONG_LONG', 3)
NVML_VALUE_TYPE_SIGNED_LONG_LONG: int = getattr(_pynvml, 'NVML_VALUE_TYPE_SIGNED_LONG_LONG', 4)
NVML_VALUE_TYPE_SIGNED_INT: int = getattr(_pynvml, 'NVML_VALUE_TYPE_SIGNED_INT', 5)
# pylint: enable=no-member

# New members in `libnvml` #########################################################################

__flags: list[int] = []
__initialized: bool = False
__lock: _threading.Lock = _threading.Lock()

LOGGER: _logging.Logger = _logging.getLogger(__name__)
try:
    LOGGER.setLevel(_os.getenv('LOGLEVEL', default='WARNING').upper())
except (ValueError, TypeError):
    pass
if not LOGGER.hasHandlers() and LOGGER.isEnabledFor(_logging.DEBUG):
    _formatter = _logging.Formatter(
        '[%(levelname)s] %(asctime)s %(name)s::%(funcName)s: %(message)s',
    )
    _stream_handler = _logging.StreamHandler()
    _stream_handler.setFormatter(_formatter)
    _file_handler = _logging.FileHandler('nvitop.log')
    _file_handler.setFormatter(_formatter)
    LOGGER.addHandler(_stream_handler)
    LOGGER.addHandler(_file_handler)
    del _formatter, _stream_handler, _file_handler

UNKNOWN_FUNCTIONS: dict[str, tuple[_Callable | str, NVMLError_FunctionNotFound]] = {}
UNKNOWN_FUNCTIONS_CACHE_SIZE: int = 1024
VERSIONED_PATTERN: _re.Pattern = _re.compile(r'^(?P<name>\w+)(?P<suffix>_v(\d)+)$')


def _lazy_init() -> None:
    """Lazily initialize the NVML context.

    Raises:
        NVMLError_LibraryNotFound:
            If cannot find the NVML library, usually the NVIDIA driver is not installed.
        NVMLError_DriverNotLoaded:
            If NVIDIA driver is not loaded.
        NVMLError_LibRmVersionMismatch:
            If RM detects a driver/library version mismatch, usually after an upgrade for NVIDIA
            driver without reloading the kernel module.
        AttributeError:
            If cannot find function :func:`pynvml.nvmlInitWithFlags`, usually the :mod:`pynvml` module
            is overridden by other modules. Need to reinstall package ``nvidia-ml-py``.
    """
    with __lock:
        if __initialized:
            return
    nvmlInit()
    _atexit.register(nvmlShutdown)


def nvmlInit() -> None:  # pylint: disable=function-redefined
    """Initialize the NVML context with default flag (0).

    Raises:
        NVMLError_LibraryNotFound:
            If cannot find the NVML library, usually the NVIDIA driver is not installed.
        NVMLError_DriverNotLoaded:
            If NVIDIA driver is not loaded.
        NVMLError_LibRmVersionMismatch:
            If RM detects a driver/library version mismatch, usually after an upgrade for NVIDIA
            driver without reloading the kernel module.
        AttributeError:
            If cannot find function :func:`pynvml.nvmlInitWithFlags`, usually the :mod:`pynvml` module
            is overridden by other modules. Need to reinstall package ``nvidia-ml-py``.
    """
    nvmlInitWithFlags(0)


def nvmlInitWithFlags(flags: int) -> None:  # pylint: disable=function-redefined
    """Initialize the NVML context with the given flags.

    Raises:
        NVMLError_LibraryNotFound:
            If cannot find the NVML library, usually the NVIDIA driver is not installed.
        NVMLError_DriverNotLoaded:
            If NVIDIA driver is not loaded.
        NVMLError_LibRmVersionMismatch:
            If RM detects a driver/library version mismatch, usually after an upgrade for NVIDIA
            driver without reloading the kernel module.
        AttributeError:
            If cannot find function :func:`pynvml.nvmlInitWithFlags`, usually the :mod:`pynvml` module
            is overridden by other modules. Need to reinstall package ``nvidia-ml-py``.
    """
    global __flags, __initialized  # pylint: disable=global-statement,global-variable-not-assigned

    with __lock:
        if len(__flags) > 0 and flags == __flags[-1]:
            __initialized = True
            return

    try:
        _pynvml.nvmlInitWithFlags(flags)
    except NVMLError_LibraryNotFound:
        message = (
            'FATAL ERROR: NVIDIA Management Library (NVML) not found.\n'
            'HINT: The NVIDIA Management Library ships with the NVIDIA display driver (available at\n'
            '      https://www.nvidia.com/Download/index.aspx), or can be downloaded as part of the\n'
            '      NVIDIA CUDA Toolkit (available at https://developer.nvidia.com/cuda-downloads).\n'
            '      The lists of OS platforms and NVIDIA-GPUs supported by the NVML library can be\n'
            '      found in the NVML API Reference at https://docs.nvidia.com/deploy/nvml-api.'
        )
        for text, color, attrs in (
            ('FATAL ERROR:', 'red', ('bold',)),
            ('HINT:', 'yellow', ('bold',)),
            ('https://www.nvidia.com/Download/index.aspx', None, ('underline',)),
            ('https://developer.nvidia.com/cuda-downloads', None, ('underline',)),
            ('https://docs.nvidia.com/deploy/nvml-api', None, ('underline',)),
        ):
            message = message.replace(text, __colored(text, color=color, attrs=attrs))

        LOGGER.critical(message)
        raise
    except AttributeError:
        message = (
            'FATAL ERROR: The dependency package `nvidia-ml-py` is corrupted. You may have installed\n'
            '             other packages overriding the module `pynvml`.\n'
            'Please reinstall `nvitop` with command:\n'
            '    python3 -m pip install --force-reinstall nvitop'
        )
        for text, color, attrs in (
            ('FATAL ERROR:', 'red', ('bold',)),
            ('nvidia-ml-py', None, ('bold',)),
            ('pynvml', None, ('bold',)),
            ('nvitop', None, ('bold',)),
        ):
            message = message.replace(text, __colored(text, color=color, attrs=attrs), 1)

        LOGGER.critical(message)
        raise

    with __lock:
        __flags.append(flags)
        __initialized = True


def nvmlShutdown() -> None:  # pylint: disable=function-redefined
    """Shutdown the NVML context.

    Raises:
        NVMLError_LibraryNotFound:
            If cannot find the NVML library, usually the NVIDIA driver is not installed.
        NVMLError_DriverNotLoaded:
            If NVIDIA driver is not loaded.
        NVMLError_LibRmVersionMismatch:
            If RM detects a driver/library version mismatch, usually after an upgrade for NVIDIA
            driver without reloading the kernel module.
        NVMLError_Uninitialized:
            If NVML was not first initialized with :func:`nvmlInit`.
    """
    global __flags, __initialized  # pylint: disable=global-statement,global-variable-not-assigned

    _pynvml.nvmlShutdown()
    with __lock:
        try:
            __flags.pop()
        except IndexError:
            pass
        __initialized = len(__flags) > 0


def nvmlQuery(
    func: _Callable[..., _Any] | str,
    *args: _Any,
    default: _Any = NA,
    ignore_errors: bool = True,
    ignore_function_not_found: bool = False,
    **kwargs: _Any,
) -> _Any:
    """Call a function with the given arguments from NVML.

    The NVML context will be automatically initialized.

    Args:
        func (Union[Callable[..., Any], str]):
            The function to call. If it is given by string, lookup for the function first from
            module :mod:`pynvml`.
        default (Any):
            The default value if the query fails.
        ignore_errors (bool):
            Whether to ignore errors and return the default value.
        ignore_function_not_found (bool):
            Whether to ignore function not found errors and return the default value. If set to
            :data:`False`, an error message will be logged to the logger.
        *args:
            Positional arguments to pass to the query function.
        **kwargs:
            Keyword arguments to pass to the query function.

    Raises:
        NVMLError_LibraryNotFound:
            If cannot find the NVML library, usually the NVIDIA driver is not installed.
        NVMLError_DriverNotLoaded:
            If NVIDIA driver is not loaded.
        NVMLError_LibRmVersionMismatch:
            If RM detects a driver/library version mismatch, usually after an upgrade for NVIDIA
            driver without reloading the kernel module.
        NVMLError_FunctionNotFound:
            If the function is not found, usually the installed ``nvidia-ml-py`` is not compatible
            with the installed NVIDIA driver.
        NVMLError_NotSupported:
            If the function is not supported by the driver or the device.
        NVMLError_InvalidArgument:
            If passed with an invalid argument.
    """
    global UNKNOWN_FUNCTIONS  # pylint: disable=global-statement,global-variable-not-assigned

    _lazy_init()

    try:
        if isinstance(func, str):
            try:
                func = getattr(__modself, func)
            except AttributeError as e1:
                raise NVMLError_FunctionNotFound from e1

        retval = func(*args, **kwargs)  # type: ignore[operator]
    except NVMLError_FunctionNotFound as e2:
        if not ignore_function_not_found:
            identifier = (
                func
                if isinstance(func, str)
                else (_inspect.getsource(func) if func.__name__ == '<lambda>' else repr(func))
            )
            with __lock:
                if (
                    identifier not in UNKNOWN_FUNCTIONS
                    and len(UNKNOWN_FUNCTIONS) < UNKNOWN_FUNCTIONS_CACHE_SIZE
                ):
                    UNKNOWN_FUNCTIONS[identifier] = (func, e2)
                    LOGGER.error(
                        (
                            'ERROR: A FunctionNotFound error occurred while calling %s.\n'
                            'Please verify whether the `nvidia-ml-py` package is '
                            'compatible with your NVIDIA driver version.'
                        ),
                        f'nvmlQuery({func!r}, *args, **kwargs)',
                    )
        if ignore_errors or ignore_function_not_found:
            return default
        raise
    except NVMLError:
        if ignore_errors:
            return default
        raise

    if isinstance(retval, bytes):
        retval = retval.decode('utf-8')
    return retval


def nvmlQueryFieldValues(
    handle: c_nvmlDevice_t,
    field_ids: list[int | tuple[int, int]],
) -> list[tuple[float | int | NaType, int]]:
    """Query multiple field values from NVML.

    Request values for a list of fields for a device. This API allows multiple fields to be queried
    at once. If any of the underlying fieldIds are populated by the same driver call, the results
    for those field IDs will be populated from a single call rather than making a driver call for
    each fieldId.

    Raises:
        NVMLError_InvalidArgument:
            If device or field_ids is invalid.
    """
    field_values = nvmlQuery('nvmlDeviceGetFieldValues', handle, field_ids)

    if not nvmlCheckReturn(field_values):
        timestamp = _time.time_ns() // 1000
        return [(NA, timestamp) for _ in range(len(field_ids))]

    values_with_timestamps: list[tuple[float | int | NaType, int]] = []
    for field_value in field_values:
        timestamp = field_value.timestamp
        if field_value.nvmlReturn != NVML_SUCCESS:
            value = NA
            timestamp = _time.time_ns() // 1000
        elif field_value.valueType == NVML_VALUE_TYPE_DOUBLE:
            value = field_value.value.dVal
        elif field_value.valueType == NVML_VALUE_TYPE_UNSIGNED_INT:
            value = field_value.value.uiVal
        elif field_value.valueType == NVML_VALUE_TYPE_UNSIGNED_LONG:
            value = field_value.value.ulVal
        elif field_value.valueType == NVML_VALUE_TYPE_UNSIGNED_LONG_LONG:
            value = field_value.value.ullVal
        elif field_value.valueType == NVML_VALUE_TYPE_SIGNED_LONG_LONG:
            value = field_value.value.llVal
        elif field_value.valueType == NVML_VALUE_TYPE_SIGNED_INT:
            value = field_value.value.iVal
        else:
            value = NA
        values_with_timestamps.append((value, timestamp))
    return values_with_timestamps


def nvmlCheckReturn(
    retval: _Any,
    types: type | tuple[type, ...] | None = None,
) -> bool:
    """Check whether the return value is not :const:`nvitop.NA` and is one of the given types."""
    if types is None:
        return retval != NA
    return retval != NA and isinstance(retval, types)


# Patch layers for backward compatibility ##########################################################
_pynvml_installation_corrupted: bool = not callable(
    getattr(_pynvml, '_nvmlGetFunctionPointer', None),
)

# Patch function `nvmlDeviceGet{Compute,Graphics,MPSCompute}RunningProcesses`
if not _pynvml_installation_corrupted:
    # pylint: disable-next=missing-class-docstring,too-few-public-methods,function-redefined
    class c_nvmlProcessInfo_v1_t(_pynvml._PrintableStructure):  # pylint: disable=protected-access
        _fields_: _ClassVar[list[tuple[str, type]]] = [
            # Process ID
            ('pid', _ctypes.c_uint),
            # Amount of used GPU memory in bytes.
            # Under WDDM, NVML_VALUE_NOT_AVAILABLE is always reported because Windows KMD manages
            # all the memory and not the NVIDIA driver.
            ('usedGpuMemory', _ctypes.c_ulonglong),
        ]
        _fmt_: _ClassVar[dict[str, str]] = {
            'usedGpuMemory': '%d B',
        }

    # pylint: disable-next=missing-class-docstring,too-few-public-methods,function-redefined
    class c_nvmlProcessInfo_v2_t(_pynvml._PrintableStructure):  # pylint: disable=protected-access
        _fields_: _ClassVar[list[tuple[str, type]]] = [
            # Process ID
            ('pid', _ctypes.c_uint),
            # Amount of used GPU memory in bytes.
            # Under WDDM, NVML_VALUE_NOT_AVAILABLE is always reported because Windows KMD manages
            # all the memory and not the NVIDIA driver.
            ('usedGpuMemory', _ctypes.c_ulonglong),
            # If MIG is enabled, stores a valid GPU instance ID. gpuInstanceId is set to 0xFFFFFFFF
            # otherwise.
            ('gpuInstanceId', _ctypes.c_uint),
            # If MIG is enabled, stores a valid compute instance ID. computeInstanceId is set to
            # 0xFFFFFFFF otherwise.
            ('computeInstanceId', _ctypes.c_uint),
        ]
        _fmt_: _ClassVar[dict[str, str]] = {
            'usedGpuMemory': '%d B',
        }

    # pylint: disable-next=missing-class-docstring,too-few-public-methods,function-redefined
    class c_nvmlProcessInfo_v3_t(_pynvml._PrintableStructure):  # pylint: disable=protected-access
        _fields_: _ClassVar[list[tuple[str, type]]] = [
            # Process ID
            ('pid', _ctypes.c_uint),
            # Amount of used GPU memory in bytes.
            # Under WDDM, NVML_VALUE_NOT_AVAILABLE is always reported because Windows KMD manages
            # all the memory and not the NVIDIA driver.
            ('usedGpuMemory', _ctypes.c_ulonglong),
            # If MIG is enabled, stores a valid GPU instance ID. gpuInstanceId is set to 0xFFFFFFFF
            # otherwise.
            ('gpuInstanceId', _ctypes.c_uint),
            # If MIG is enabled, stores a valid compute instance ID. computeInstanceId is set to
            # 0xFFFFFFFF otherwise.
            ('computeInstanceId', _ctypes.c_uint),
            # Amount of used GPU conf compute protected memory in bytes.
            ('usedGpuCcProtectedMemory', _ctypes.c_ulonglong),
        ]
        _fmt_: _ClassVar[dict[str, str]] = {
            'usedGpuMemory': '%d B',
            'usedGpuCcProtectedMemory': '%d B',
        }

    __get_running_processes_version_suffix = None
    c_nvmlProcessInfo_t = c_nvmlProcessInfo_v3_t

    def __determine_get_running_processes_version_suffix() -> str:
        global __get_running_processes_version_suffix, c_nvmlProcessInfo_t  # pylint: disable=global-statement

        if __get_running_processes_version_suffix is None:
            # pylint: disable-next=protected-access,no-member
            _nvmlGetFunctionPointer = _pynvml._nvmlGetFunctionPointer
            __get_running_processes_version_suffix = '_v3'

            def lookup(symbol: str) -> _Any | None:
                try:
                    ptr = _nvmlGetFunctionPointer(symbol)
                except NVMLError_FunctionNotFound:
                    LOGGER.debug('Failed to found symbol `%s`.', symbol)
                    return None
                LOGGER.debug('Found symbol `%s`.', symbol)
                return ptr

            if lookup('nvmlDeviceGetComputeRunningProcesses_v3'):
                if lookup('nvmlDeviceGetConfComputeMemSizeInfo') and not lookup(
                    'nvmlDeviceGetRunningProcessDetailList',
                ):
                    LOGGER.debug(
                        'NVML get running process version 3 API with v3 type struct is available.',
                    )
                else:
                    c_nvmlProcessInfo_t = c_nvmlProcessInfo_v2_t
                    LOGGER.debug(
                        'NVML get running process version 3 API with v3 type struct is not '
                        'available due to incompatible NVIDIA driver. Fallback to use get running '
                        'process version 3 API with v2 type struct.',
                    )
            else:
                c_nvmlProcessInfo_t = c_nvmlProcessInfo_v2_t
                __get_running_processes_version_suffix = '_v2'
                LOGGER.debug(
                    'NVML get running process version 3 API with v3 type struct is not available '
                    'due to incompatible NVIDIA driver. Fallback to use get running process '
                    'version 2 API with v2 type struct.',
                )
                if lookup('nvmlDeviceGetComputeRunningProcesses_v2'):
                    LOGGER.debug(
                        'NVML get running process version 2 API with v2 type struct is available.',
                    )
                else:
                    c_nvmlProcessInfo_t = c_nvmlProcessInfo_v1_t
                    __get_running_processes_version_suffix = ''
                    LOGGER.debug(
                        'NVML get running process version 2 API with v2 type struct is not '
                        'available due to incompatible NVIDIA driver. Fallback to use get '
                        'running process version 1 API with v1 type struct.',
                    )

        return __get_running_processes_version_suffix

    def __nvml_device_get_running_processes(
        func: str,
        handle: c_nvmlDevice_t,
    ) -> list[c_nvmlProcessInfo_t]:
        """Helper function for :func:`nvmlDeviceGet{Compute,Graphics,MPSCompute}RunningProcesses`.

        Modified from function :func:`pynvml.nvmlDeviceGetComputeRunningProcesses` in package
        `nvidia-ml-py <https://pypi.org/project/nvidia-ml-py>`_.
        """
        version_suffix = __determine_get_running_processes_version_suffix()

        # First call to get the size
        c_count = _ctypes.c_uint(0)
        # pylint: disable-next=protected-access
        fn = _pynvml._nvmlGetFunctionPointer(f'{func}{version_suffix}')
        ret = fn(handle, _ctypes.byref(c_count), None)

        if ret == NVML_SUCCESS:
            # Special case, no running processes
            return []
        if ret == NVML_ERROR_INSUFFICIENT_SIZE:
            # Typical case
            # Oversize the array in case more processes are created
            c_count.value = c_count.value * 2 + 5
            process_array = c_nvmlProcessInfo_t * c_count.value  # type: ignore[operator]
            c_processes = process_array()  # type: ignore[operator]

            # Make the call again
            ret = fn(handle, _ctypes.byref(c_count), c_processes)
            _pynvml._nvmlCheckReturn(ret)  # pylint: disable=protected-access

            processes = []
            for i in range(c_count.value):
                # Use an alternative struct for this object
                obj = _pynvml.nvmlStructToFriendlyObject(c_processes[i])
                if obj.usedGpuMemory == ULONGLONG_MAX:
                    # Special case for WDDM on Windows, see comment above
                    obj.usedGpuMemory = None
                processes.append(obj)

            return processes

        # Error case
        raise NVMLError(ret)

    def nvmlDeviceGetComputeRunningProcesses(  # pylint: disable=function-redefined
        handle: c_nvmlDevice_t,
    ) -> list[c_nvmlProcessInfo_t]:
        """Get information about processes with a compute context on a device.

        Note:
            - In MIG mode, if device handle is provided, the API returns aggregate information, only
              if the caller has appropriate privileges. Per-instance information can be queried by
              using specific MIG device handles.

        Raises:
            NVMLError_Uninitialized:
                If NVML was not first initialized with :func:`nvmlInit`.
            NVMLError_NoPermission:
                If the user doesn't have permission to perform this operation.
            NVMLError_InvalidArgument:
                If device is invalid.
            NVMLError_GpuIsLost:
                If the target GPU has fallen off the bus or is otherwise inaccessible.
            NVMLError_Unknown:
                On any unexpected error.
        """
        return __nvml_device_get_running_processes(
            'nvmlDeviceGetComputeRunningProcesses',
            handle,
        )

    def nvmlDeviceGetGraphicsRunningProcesses(  # pylint: disable=function-redefined
        handle: c_nvmlDevice_t,
    ) -> list[c_nvmlProcessInfo_t]:
        """Get information about processes with a graphics context on a device.

        Note:
            - In MIG mode, if device handle is provided, the API returns aggregate information, only
              if the caller has appropriate privileges. Per-instance information can be queried by
              using specific MIG device handles.

        Raises:
            NVMLError_Uninitialized:
                If NVML was not first initialized with :func:`nvmlInit`.
            NVMLError_NoPermission:
                If the user doesn't have permission to perform this operation.
            NVMLError_InvalidArgument:
                If device is invalid.
            NVMLError_GpuIsLost:
                If the target GPU has fallen off the bus or is otherwise inaccessible.
            NVMLError_Unknown:
                On any unexpected error.
        """
        return __nvml_device_get_running_processes(
            'nvmlDeviceGetGraphicsRunningProcesses',
            handle,
        )

    def nvmlDeviceGetMPSComputeRunningProcesses(  # pylint: disable=function-redefined
        handle: c_nvmlDevice_t,
    ) -> list[c_nvmlProcessInfo_t]:
        """Get information about processes with a MPS compute context on a device.

        Note:
            - In MIG mode, if device handle is provided, the API returns aggregate information, only
              if the caller has appropriate privileges. Per-instance information can be queried by
              using specific MIG device handles.

        Raises:
            NVMLError_Uninitialized:
                If NVML was not first initialized with :func:`nvmlInit`.
            NVMLError_NoPermission:
                If the user doesn't have permission to perform this operation.
            NVMLError_InvalidArgument:
                If device is invalid.
            NVMLError_GpuIsLost:
                If the target GPU has fallen off the bus or is otherwise inaccessible.
            NVMLError_Unknown:
                On any unexpected error.
        """
        return __nvml_device_get_running_processes(
            'nvmlDeviceGetMPSComputeRunningProcesses',
            handle,
        )

else:
    LOGGER.warning(
        'Your installed package `nvidia-ml-py` is corrupted. '
        'Skip patch functions `nvmlDeviceGet{Compute,Graphics,MPSCompute}RunningProcesses`. '
        'You may get incorrect or incomplete results. Please consider reinstall package '
        '`nvidia-ml-py` via `pip3 install --force-reinstall nvidia-ml-py nvitop`.',
    )

# Patch function `nvmlDeviceGetMemoryInfo`
if not _pynvml_installation_corrupted:
    # pylint: disable-next=missing-class-docstring,too-few-public-methods,function-redefined
    class c_nvmlMemory_v1_t(_pynvml._PrintableStructure):  # pylint: disable=protected-access
        _fields_: _ClassVar[list[tuple[str, type]]] = [
            # Total physical device memory (in bytes).
            ('total', _ctypes.c_ulonglong),
            # Unallocated device memory (in bytes).
            ('free', _ctypes.c_ulonglong),
            # Allocated device memory (in bytes).
            # Note that the driver/GPU always sets aside a small amount of memory for bookkeeping.
            ('used', _ctypes.c_ulonglong),
        ]
        _fmt_: _ClassVar[dict[str, str]] = {'<default>': '%d B'}

    # pylint: disable-next=missing-class-docstring,too-few-public-methods,function-redefined
    class c_nvmlMemory_v2_t(_pynvml._PrintableStructure):  # pylint: disable=protected-access
        _fields_: _ClassVar[list[tuple[str, type]]] = [
            # Structure format version (must be 2).
            ('version', _ctypes.c_uint),
            # Total physical device memory (in bytes).
            ('total', _ctypes.c_ulonglong),
            # Device memory (in bytes) reserved for system use (driver or firmware).
            ('reserved', _ctypes.c_ulonglong),
            # Unallocated device memory (in bytes).
            ('free', _ctypes.c_ulonglong),
            # Allocated device memory (in bytes).
            # Note that the driver/GPU always sets aside a small amount of memory for bookkeeping.
            ('used', _ctypes.c_ulonglong),
        ]
        _fmt_: _ClassVar[dict[str, str]] = {'<default>': '%d B'}

    nvmlMemory_v2 = getattr(_pynvml, 'nvmlMemory_v2', _ctypes.sizeof(c_nvmlMemory_v2_t) | 2 << 24)
    __get_memory_info_version_suffix = None
    c_nvmlMemory_t = c_nvmlMemory_v2_t

    def __determine_get_memory_info_version_suffix() -> str:
        global __get_memory_info_version_suffix, c_nvmlMemory_t  # pylint: disable=global-statement

        if __get_memory_info_version_suffix is None:
            # pylint: disable-next=protected-access,no-member
            _nvmlGetFunctionPointer = _pynvml._nvmlGetFunctionPointer
            __get_memory_info_version_suffix = '_v2'
            try:
                _nvmlGetFunctionPointer('nvmlDeviceGetMemoryInfo_v2')
            except NVMLError_FunctionNotFound:
                LOGGER.debug('Failed to found symbol `nvmlDeviceGetMemoryInfo_v2`.')
                c_nvmlMemory_t = c_nvmlMemory_v1_t
                __get_memory_info_version_suffix = ''
                LOGGER.debug(
                    'NVML get memory info version 2 API is not available due to incompatible '
                    'NVIDIA driver. Fallback to use NVML get memory info version 1 API.',
                )
            else:
                LOGGER.debug('Found symbol `nvmlDeviceGetMemoryInfo_v2`.')
                LOGGER.debug('NVML get memory info version 2 is available.')

        return __get_memory_info_version_suffix

    def nvmlDeviceGetMemoryInfo(  # pylint: disable=function-redefined
        handle: c_nvmlDevice_t,
    ) -> c_nvmlMemory_t:
        """Retrieve the amount of used, free, reserved and total memory available on the device, in bytes.

        Note:
            - The version 2 API adds additional memory information. The reserved amount is supported
              on version 2 only.
            - In MIG mode, if device handle is provided, the API returns aggregate information, only
              if the caller has appropriate privileges. Per-instance information can be queried by
              using specific MIG device handles.

        Raises:
            NVMLError_Uninitialized:
                If NVML was not first initialized with :func:`nvmlInit`.
            NVMLError_NoPermission:
                If the user doesn't have permission to perform this operation.
            NVMLError_InvalidArgument:
                If device is invalid.
            NVMLError_GpuIsLost:
                If the target GPU has fallen off the bus or is otherwise inaccessible.
            NVMLError_Unknown:
                On any unexpected error.
        """
        version_suffix = __determine_get_memory_info_version_suffix()
        if version_suffix == '_v2':
            c_memory = c_nvmlMemory_v2_t()
            c_memory.version = nvmlMemory_v2  # pylint: disable=attribute-defined-outside-init
            # pylint: disable-next=protected-access
            fn = _pynvml._nvmlGetFunctionPointer('nvmlDeviceGetMemoryInfo_v2')
        elif version_suffix in {'_v1', ''}:
            c_memory = c_nvmlMemory_v1_t()
            # pylint: disable-next=protected-access
            fn = _pynvml._nvmlGetFunctionPointer('nvmlDeviceGetMemoryInfo')
        else:
            raise ValueError(
                f'Unknown version suffix {version_suffix!r} for '
                'function `nvmlDeviceGetMemoryInfo`.',
            )
        ret = fn(handle, _ctypes.byref(c_memory))
        _pynvml._nvmlCheckReturn(ret)  # pylint: disable=protected-access
        return c_memory

else:
    LOGGER.warning(
        'Your installed package `nvidia-ml-py` is corrupted. '
        'Skip patch functions `nvmlDeviceGetMemoryInfo`. '
        'You may get incorrect or incomplete results. Please consider reinstall package '
        '`nvidia-ml-py` via `pip3 install --force-reinstall nvidia-ml-py nvitop`.',
    )


# Add support for lookup fallback and context manager ##############################################
class _CustomModule(_ModuleType):
    """Modified module type to support lookup fallback and context manager.

    Automatic lookup fallback:

        >>> libnvml.c_nvmlGpuInstance_t  # fallback to pynvml.c_nvmlGpuInstance_t
        <class 'pynvml.LP_struct_c_nvmlGpuInstance_t'>

    Context manager:

        >>> with libnvml:
        ...     handle = libnvml.nvmlDeviceGetHandleByIndex(0)
        ... # The NVML context has been shutdown
    """

    def __getattribute__(self, name: str) -> _Any | _Callable[..., _Any]:
        """Get a member from the current module. Fallback to the original package if missing."""
        try:
            return super().__getattribute__(name)
        except AttributeError:
            return getattr(_pynvml, name)

    def __enter__(self) -> _Self:
        """Entry of the context manager for ``with`` statement."""
        _lazy_init()
        return self

    def __exit__(self, *exc: object) -> None:
        """Shutdown the NVML context in the context manager for ``with`` statement."""
        try:
            nvmlShutdown()
        except NVMLError:
            pass


# Replace entry in sys.modules for this module with an instance of _CustomModule
__modself = _sys.modules[__name__]
__modself.__class__ = _CustomModule



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

from collections import OrderedDict
from functools import partial
from itertools import islice

from nvitop.gui.library import Displayable, GpuProcess, HostProcess, WideString, host


class EnvironScreen(Displayable):  # pylint: disable=too-many-instance-attributes
    NAME = 'environ'

    def __init__(self, win, root):
        super().__init__(win, root)

        self.this = HostProcess()

        self._process = None
        self._environ = None
        self.items = None
        self.username = None
        self.command = None

        self.x_offset = 0
        self._y_offset = 0
        self.scroll_offset = 0
        self.y_mouse = None

        self._height = 0
        self.x, self.y = root.x, root.y
        self.width, self.height = root.width, root.height

    @property
    def process(self):
        return self._process

    @process.setter
    def process(self, value):
        if value is None:
            value = self.this

        self._process = value

        with self.process.oneshot():
            try:
                self.environ = self.process.environ().copy()
            except host.PsutilError:
                self.environ = None

            try:
                self.command = self.process.command()
            except host.PsutilError:
                self.command = 'N/A'

            try:
                self.username = self.process.username()
            except host.PsutilError:
                self.username = 'N/A'

        self.command = WideString(self.command)
        self.username = WideString(self.username)

    @property
    def environ(self):
        return self._environ

    @environ.setter
    def environ(self, value):
        newline = '␤' if not self.root.ascii else '?'

        def normalize(s):
            return s.replace('\n', newline)

        if value is not None:
            self.items = [
                (WideString(key), WideString(f'{key}={normalize(value[key])}'))
                for key in sorted(value.keys())
            ]
            value = OrderedDict(self.items)
        else:
            self.items = None
        self._environ = value
        self.x_offset = 0
        self.y_offset = 0
        self.scroll_offset = 0

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, value):
        self._height = value
        try:
            self.y_offset = self.y_offset
        except AttributeError:
            pass

    @property
    def display_height(self):
        return self.height - 2

    @property
    def y_offset(self):
        return self._y_offset

    @y_offset.setter
    def y_offset(self, value):
        if self.environ is None:
            self._y_offset = 0
            self.scroll_offset = 0
            return

        n_items = len(self.environ)
        self._y_offset = max(0, min(value, n_items - 1))
        if n_items <= self.scroll_offset + self.display_height:
            self.scroll_offset = max(0, n_items - self.display_height)
        elif self.y_offset > self.scroll_offset + self.display_height - 1:
            self.scroll_offset = self.y_offset - self.display_height + 1
        self.scroll_offset = min(self.scroll_offset, self.y_offset)

    def move(self, direction, wheel=False):
        if self.environ is not None and wheel:
            n_items = len(self.environ)
            old_scroll_offset = self.scroll_offset
            self.scroll_offset = max(
                0,
                min(self.scroll_offset + direction, n_items - self.display_height),
            )
            direction -= self.scroll_offset - old_scroll_offset
            self._y_offset += self.scroll_offset - old_scroll_offset
        self.y_offset = self.y_offset + direction

    def update_size(self, termsize=None):
        n_term_lines, n_term_cols = termsize = super().update_size(termsize=termsize)

        self.width = n_term_cols - self.x
        self.height = n_term_lines - self.y

        return termsize

    def draw(self):
        self.color_reset()

        if isinstance(self.process, GpuProcess):
            process_type = 'GPU: ' + self.process.type.replace('C', 'Compute').replace(
                'G',
                'Graphics',
            )
        else:
            process_type = 'Host'
        header_prefix = WideString(
            f'Environment of process {self.process.pid} ({self.username}@{process_type}): ',
        )
        offset = max(0, min(self.x_offset, len(self.command) + len(header_prefix) - self.width))
        header = str((header_prefix + self.command[offset:]).ljust(self.width)[: self.width])

        self.addstr(self.y, self.x, header)
        self.addstr(self.y + 1, self.x, '#' * self.width)
        self.color_at(self.y, self.x, width=len(header_prefix) - 1, fg='cyan', attr='bold')
        self.color_at(self.y + 1, self.x, width=self.width, fg='green', attr='bold')

        if self.environ is None:
            self.addstr(self.y + 2, self.x, 'Could not read process environment.')
            self.color_at(self.y + 2, self.x, width=self.width, fg='cyan', attr='reverse')
            return

        items = islice(self.items, self.scroll_offset, self.scroll_offset + self.display_height)
        for y, (key, line) in enumerate(items, start=self.y + 2):
            key_length = len(key)
            line = str(line[self.x_offset :].ljust(self.width)[: self.width])
            self.addstr(y, self.x, line)
            if self.x_offset < key_length:
                self.color_at(y, self.x, width=key_length - self.x_offset, fg='blue', attr='bold')
            if self.x_offset < key_length + 1:
                self.color_at(y, self.x + key_length - self.x_offset, width=1, fg='magenta')

            if y == self.y_mouse:
                self.y_offset = y - (self.y + 2 - self.scroll_offset)

            if y == self.y + 2 - self.scroll_offset + self.y_offset:
                self.color_at(y, self.x, width=self.width, fg='cyan', attr='bold | reverse')

    def finalize(self):
        self.y_mouse = None
        super().finalize()

    def press(self, key):
        self.root.keymaps.use_keymap('environ')
        self.root.press(key)

    def click(self, event):
        if event.pressed(1) or event.pressed(3) or event.clicked(1) or event.clicked(3):
            self.y_mouse = event.y
            return True

        direction = event.wheel_direction()
        if event.shift():
            self.x_offset = max(0, self.x_offset + 2 * direction)
        else:
            self.move(direction=direction, wheel=True)
        return True

    def init_keybindings(self):
        def refresh_environ():
            self.process = self.root.previous_screen.selection.process
            self.need_redraw = True

        def environ_left():
            self.x_offset = max(0, self.x_offset - 5)

        def environ_right():
            self.x_offset += 5

        def environ_begin():
            self.x_offset = 0

        def environ_move(direction):
            self.move(direction=direction)

        keymaps = self.root.keymaps

        keymaps.bind('environ', 'r', refresh_environ)
        keymaps.copy('environ', 'r', 'R')
        keymaps.copy('environ', 'r', '<C-r>')
        keymaps.copy('environ', 'r', '<F5>')
        keymaps.bind('environ', '<Left>', environ_left)
        keymaps.copy('environ', '<Left>', '<A-h>')
        keymaps.bind('environ', '<Right>', environ_right)
        keymaps.copy('environ', '<Right>', '<A-l>')
        keymaps.bind('environ', '<C-a>', environ_begin)
        keymaps.copy('environ', '<C-a>', '^')
        keymaps.bind('environ', '<Up>', partial(environ_move, direction=-1))
        keymaps.copy('environ', '<Up>', '<S-Tab>')
        keymaps.copy('environ', '<Up>', '<A-k>')
        keymaps.copy('environ', '<Up>', '<PageUp>')
        keymaps.copy('environ', '<Up>', '[')
        keymaps.bind('environ', '<Down>', partial(environ_move, direction=+1))
        keymaps.copy('environ', '<Down>', '<Tab>')
        keymaps.copy('environ', '<Down>', '<A-j>')
        keymaps.copy('environ', '<Down>', '<PageDown>')
        keymaps.copy('environ', '<Down>', ']')
        keymaps.bind('environ', '<Home>', partial(environ_move, direction=-(1 << 20)))
        keymaps.bind('environ', '<End>', partial(environ_move, direction=+(1 << 20)))



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
#
# Copyright 2021-2023 Xuehai Pan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""An interactive NVIDIA-GPU process viewer and beyond, the one-stop solution for GPU process management."""

import sys

from nvitop import api
from nvitop.api import *  # noqa: F403
from nvitop.api import collector, device, host, libcuda, libcudart, libnvml, process, utils
from nvitop.select import select_devices
from nvitop.version import __version__


__all__ = [*api.__all__, 'select_devices']

# Add submodules to the top-level namespace
for submodule in (collector, device, host, libcuda, libcudart, libnvml, process, utils):
    sys.modules[f'{__name__}.{submodule.__name__.rpartition(".")[-1]}'] = submodule

# Remove the nvitop.select module from sys.modules
# Required for `python -m nvitop.select` to work properly
sys.modules.pop(f'{__name__}.select', None)

del sys



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import functools
import itertools
import threading
import time
from collections import deque

from nvitop.api import NA


BOUND_UPDATE_INTERVAL = 1.0

# fmt: off
VALUE2SYMBOL_UP = {
    (0, 0): ' ', (0, 1): '⢀', (0, 2): '⢠', (0, 3): '⢰', (0, 4): '⢸',
    (1, 0): '⡀', (1, 1): '⣀', (1, 2): '⣠', (1, 3): '⣰', (1, 4): '⣸',
    (2, 0): '⡄', (2, 1): '⣄', (2, 2): '⣤', (2, 3): '⣴', (2, 4): '⣼',
    (3, 0): '⡆', (3, 1): '⣆', (3, 2): '⣦', (3, 3): '⣶', (3, 4): '⣾',
    (4, 0): '⡇', (4, 1): '⣇', (4, 2): '⣧', (4, 3): '⣷', (4, 4): '⣿',
}
VALUE2SYMBOL_DOWN = {
    (0, 0): ' ', (0, 1): '⠈', (0, 2): '⠘', (0, 3): '⠸', (0, 4): '⢸',
    (1, 0): '⠁', (1, 1): '⠉', (1, 2): '⠙', (1, 3): '⠹', (1, 4): '⢹',
    (2, 0): '⠃', (2, 1): '⠋', (2, 2): '⠛', (2, 3): '⠻', (2, 4): '⢻',
    (3, 0): '⠇', (3, 1): '⠏', (3, 2): '⠟', (3, 3): '⠿', (3, 4): '⢿',
    (4, 0): '⡇', (4, 1): '⡏', (4, 2): '⡟', (4, 3): '⡿', (4, 4): '⣿',
}
# fmt: on
SYMBOL2VALUE_UP = {v: k for k, v in VALUE2SYMBOL_UP.items()}
SYMBOL2VALUE_DOWN = {v: k for k, v in VALUE2SYMBOL_DOWN.items()}
PAIR2SYMBOL_UP = {
    (s1, s2): VALUE2SYMBOL_UP[(SYMBOL2VALUE_UP[s1][-1], SYMBOL2VALUE_UP[s2][0])]
    for s1, s2 in itertools.product(SYMBOL2VALUE_UP, repeat=2)
}
PAIR2SYMBOL_DOWN = {
    (s1, s2): VALUE2SYMBOL_DOWN[(SYMBOL2VALUE_DOWN[s1][-1], SYMBOL2VALUE_DOWN[s2][0])]
    for s1, s2 in itertools.product(SYMBOL2VALUE_DOWN, repeat=2)
}
GRAPH_SYMBOLS = ''.join(
    sorted(set(itertools.chain(VALUE2SYMBOL_UP.values(), VALUE2SYMBOL_DOWN.values()))),
).replace(' ', '')


def grouped(iterable, size, fillvalue=None):
    yield from itertools.zip_longest(*([iter(iterable)] * size), fillvalue=fillvalue)


class HistoryGraph:  # pylint: disable=too-many-instance-attributes
    MAX_WIDTH = 1024

    # pylint: disable-next=too-many-arguments
    def __init__(
        self,
        upperbound,
        width,
        height,
        format='{:.1f}'.format,  # pylint: disable=redefined-builtin
        max_format=None,
        baseline=0.0,
        dynamic_bound=False,
        min_bound=None,
        init_bound=None,
        upsidedown=False,
    ):
        assert baseline < upperbound

        self.format = format
        if max_format is None:
            max_format = format
        self.max_format = max_format

        if dynamic_bound:
            if min_bound is None:
                min_bound = baseline + 0.1 * (upperbound - baseline)
            if init_bound is None:
                init_bound = upperbound
        else:
            assert min_bound is None
            assert init_bound is None
            min_bound = init_bound = upperbound
        self.baseline = baseline
        self.min_bound = min_bound
        self.max_bound = upperbound
        self.bound = init_bound
        self.next_bound_update_at = time.monotonic()
        self._width = width
        self._height = height

        self.maxlen = 2 * self.width + 1
        self.history = deque(
            [self.baseline - 0.1] * (2 * self.MAX_WIDTH + 1),
            maxlen=(2 * self.MAX_WIDTH + 1),
        )
        self.reversed_history = deque([self.baseline - 0.1] * self.maxlen, maxlen=self.maxlen)
        self._max_value_maintainer = deque([self.baseline - 0.1] * self.maxlen, maxlen=self.maxlen)
        self.last_retval = None

        self.graph = []
        self.last_graph = []
        self.upsidedown = upsidedown
        if upsidedown:
            self.value2symbol = VALUE2SYMBOL_DOWN
            self.pair2symbol = PAIR2SYMBOL_DOWN
        else:
            self.value2symbol = VALUE2SYMBOL_UP
            self.pair2symbol = PAIR2SYMBOL_UP

        self.write_lock = threading.Lock()
        self.remake_lock = threading.Lock()
        self.remake_graph()

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        if self._width != value:
            assert isinstance(value, int)
            assert value >= 1
            self._width = value
            with self.write_lock:
                self.maxlen = 2 * self.width + 1
                self.reversed_history = deque(
                    (self.baseline - 0.1,) * self.maxlen,
                    maxlen=self.maxlen,
                )
                self._max_value_maintainer = deque(
                    (self.baseline - 0.1,) * self.maxlen,
                    maxlen=self.maxlen,
                )
                for history in itertools.islice(
                    self.history,
                    max(0, self.history.maxlen - self.maxlen),
                    self.history.maxlen,
                ):
                    if self.reversed_history[-1] == self._max_value_maintainer[0]:
                        self._max_value_maintainer.popleft()
                    while (
                        len(self._max_value_maintainer) > 0
                        and self._max_value_maintainer[-1] < history
                    ):
                        self._max_value_maintainer.pop()
                    self.reversed_history.appendleft(history)
                    self._max_value_maintainer.append(history)
                self.remake_graph()

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, value):
        if self._height != value:
            assert isinstance(value, int)
            assert value >= 1
            self._height = value
            self.remake_graph()

    @property
    def graph_size(self):
        return (self.width, self.height)

    @graph_size.setter
    def graph_size(self, value):
        width, height = value
        assert isinstance(width, int)
        assert width >= 1
        assert isinstance(height, int)
        assert height >= 1
        self._height = height
        self._width = width - 1  # trigger force remake
        self.width = width

    @property
    def last_value(self):
        return self.reversed_history[0]

    @property
    def max_value(self):
        return self._max_value_maintainer[0]

    def last_value_string(self):
        last_value = self.last_value
        if last_value >= self.baseline:
            return self.format(last_value)
        try:
            return self.format(NA)
        except ValueError:
            return NA

    __str__ = last_value_string

    def max_value_string(self):
        max_value = self.max_value
        if max_value >= self.baseline:
            return self.max_format(max_value)
        try:
            return self.max_format(NA)
        except ValueError:
            return NA

    def add(self, value):
        if value is NA:
            value = self.baseline - 0.1
        if not isinstance(value, (int, float)):
            return

        with self.write_lock:
            if self.reversed_history[-1] == self._max_value_maintainer[0]:
                self._max_value_maintainer.popleft()
            while len(self._max_value_maintainer) > 0 and self._max_value_maintainer[-1] < value:
                self._max_value_maintainer.pop()
            self.reversed_history.appendleft(value)
            self._max_value_maintainer.append(value)
            self.history.append(value)

            new_bound = self.baseline + 1.25 * (self.max_value - self.baseline)
            new_bound = min(max(new_bound, self.min_bound), self.max_bound)
            timestamp = time.monotonic()
            if new_bound != self.bound and self.next_bound_update_at <= timestamp:
                self.bound = new_bound
                self.remake_graph()
                self.next_bound_update_at = timestamp + BOUND_UPDATE_INTERVAL
                return

            self.graph, self.last_graph = self.last_graph, self.graph
            bar = self.make_bar(self.reversed_history[1], value)  # pylint: disable=disallowed-name
            for i, (line, char) in enumerate(zip(self.graph, bar)):
                self.graph[i] = (line + char)[-self.width :]

    def remake_graph(self):
        with self.remake_lock:
            if self.max_value >= self.baseline:
                reversed_bars = []
                for _, (value2, value1) in zip(
                    range(self.width),
                    grouped(self.reversed_history, size=2, fillvalue=self.baseline),
                ):
                    reversed_bars.append(self.make_bar(value1, value2))
                graph = list(map(''.join, zip(*reversed(reversed_bars))))

                for i, line in enumerate(graph):
                    graph[i] = line.rjust(self.width)[-self.width :]

                self.graph = graph
                self.last_graph = list(map(self.shift_line, self.graph))
            else:
                self.graph = [' ' * self.width for _ in range(self.height)]
                self.last_graph = [' ' * (self.width - 1) for _ in range(self.height)]

    def make_bar(self, value1, value2):
        if self.bound <= self.baseline:
            return [' '] * self.height

        value1 = self.height * min((value1 - self.baseline) / (self.bound - self.baseline), 1.0)
        value2 = self.height * min((value2 - self.baseline) / (self.bound - self.baseline), 1.0)
        if value1 >= 0.0:
            value1 = max(value1, 0.2)
        if value2 >= 0.0:
            value2 = max(value2, 0.2)
        # pylint: disable=disallowed-name,invalid-name
        bar = []
        for h in range(self.height):
            s1 = min(max(round(5 * (value1 - h)), 0), 4)
            s2 = min(max(round(5 * (value2 - h)), 0), 4)
            bar.append(self.value2symbol[(s1, s2)])
        if not self.upsidedown:
            bar.reverse()
        return bar

    def shift_line(self, line):
        return ''.join(map(self.pair2symbol.get, zip(line, line[1:])))

    def __getitem__(self, item):
        return self.reversed_history[item]

    def hook(self, func, get_value=None):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            self.last_retval = retval = value = func(*args, **kwargs)
            if get_value is not None:
                value = get_value(retval)
            self.add(value)
            return retval

        wrapped.history = self
        return wrapped

    __call__ = hook


class BufferedHistoryGraph(HistoryGraph):
    # pylint: disable-next=too-many-arguments
    def __init__(
        self,
        upperbound,
        width,
        height,
        format='{:.1f}'.format,  # pylint: disable=redefined-builtin
        max_format=None,
        baseline=0.0,
        dynamic_bound=False,
        upsidedown=False,
        min_bound=None,
        init_bound=None,
        interval=1.0,
    ):
        assert interval > 0.0
        super().__init__(
            upperbound,
            width,
            height,
            format=format,
            max_format=max_format,
            baseline=baseline,
            dynamic_bound=dynamic_bound,
            min_bound=min_bound,
            init_bound=init_bound,
            upsidedown=upsidedown,
        )

        self.interval = interval
        self.start_time = time.monotonic()
        self.last_update_time = self.start_time
        self.buffer = []

    @property
    def last_value(self):
        last_value = super().last_value
        if last_value < self.baseline and len(self.buffer) > 0:
            return sum(self.buffer) / len(self.buffer)
        return last_value

    def add(self, value):
        if value is NA:
            value = self.baseline - 0.1
        if not isinstance(value, (int, float)):
            return

        timestamp = time.monotonic()
        timedelta = timestamp - self.last_update_time
        if len(self.buffer) > 0 and timedelta >= self.interval:
            new_value = sum(self.buffer) / len(self.buffer)
            self.buffer.clear()
            last_value = self.reversed_history[0]
            if last_value >= self.baseline:
                n_interval = int(timedelta / self.interval)
                for i in range(1, n_interval):
                    super().add(last_value + (i / n_interval) * (new_value - last_value))
            super().add(new_value)

            self.last_update_time += (timedelta // self.interval) * self.interval
        self.buffer.append(value)



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

"""The interactive NVIDIA-GPU process viewer."""

import argparse
import curses
import os
import sys

from nvitop.api import HostProcess, libnvml
from nvitop.gui import UI, USERNAME, Device, colored, libcurses, set_color, setlocale_utf8
from nvitop.version import __version__


TTY = sys.stdin.isatty() and sys.stdout.isatty()
NVITOP_MONITOR_MODE = set(
    map(
        str.strip,
        os.environ.get('NVITOP_MONITOR_MODE', '').lower().split(','),
    ),
)


# pylint: disable=too-many-branches,too-many-statements
def parse_arguments() -> argparse.Namespace:
    """Parse command-line arguments for ``nvitop``."""
    coloring_rules = '{} < th1 %% <= {} < th2 %% <= {}'.format(
        colored('light', 'green'),
        colored('moderate', 'yellow'),
        colored('heavy', 'red'),
    )

    def posfloat(argstring: str) -> float:
        num = float(argstring)
        if num <= 0:
            raise ValueError
        return num

    posfloat.__name__ = 'positive float'

    parser = argparse.ArgumentParser(
        prog='nvitop',
        description='An interactive NVIDIA-GPU process viewer.',
        formatter_class=argparse.RawTextHelpFormatter,
        add_help=False,
    )
    parser.add_argument(
        '--help',
        '-h',
        dest='help',
        action='help',
        default=argparse.SUPPRESS,
        help='Show this help message and exit.',
    )
    parser.add_argument(
        '--version',
        '-V',
        dest='version',
        action='version',
        version=f'%(prog)s {__version__}',
        help="Show %(prog)s's version number and exit.",
    )

    mode = parser.add_mutually_exclusive_group()
    mode.add_argument(
        '--once',
        '-1',
        dest='once',
        action='store_true',
        help='Report query data only once.',
    )
    mode.add_argument(
        '--monitor',
        '-m',
        dest='monitor',
        type=str,
        default=argparse.SUPPRESS,
        nargs='?',
        choices=['auto', 'full', 'compact'],
        help=(
            'Run as a resource monitor. Continuously report query data and handle user inputs.\n'
            'If the argument is omitted, the value from `NVITOP_MONITOR_MODE` will be used.\n'
            '(default fallback mode: auto)'
        ),
    )

    parser.add_argument(
        '--interval',
        dest='interval',
        type=posfloat,
        default=None,
        metavar='SEC',
        help='Process status update interval in seconds. (default: 2)',
    )
    parser.add_argument(
        '--ascii',
        '--no-unicode',
        '-U',
        dest='ascii',
        action='store_true',
        help='Use ASCII characters only, which is useful for terminals without Unicode support.',
    )

    coloring = parser.add_argument_group('coloring')
    coloring.add_argument(
        '--colorful',
        dest='colorful',
        action='store_true',
        help=(
            'Use gradient colors to get spectrum-like bar charts. This option is only available\n'
            'when the terminal supports 256 colors. You may need to set environment variable\n'
            '`TERM="xterm-256color"`. Note that the terminal multiplexer, such as `tmux`, may\n'
            'override the `TREM` variable.'
        ),
    )
    coloring.add_argument(
        '--force-color',
        dest='force_color',
        action='store_true',
        help='Force colorize even when `stdout` is not a TTY terminal.',
    )
    coloring.add_argument(
        '--light',
        action='store_true',
        help=(
            'Tweak visual results for light theme terminals in monitor mode.\n'
            'Set variable `NVITOP_MONITOR_MODE="light"` on light terminals for convenience.'
        ),
    )
    gpu_thresholds = Device.GPU_UTILIZATION_THRESHOLDS
    coloring.add_argument(
        '--gpu-util-thresh',
        type=int,
        nargs=2,
        choices=range(1, 100),
        metavar=('th1', 'th2'),
        help=(
            'Thresholds of GPU utilization to determine the load intensity.\n'
            'Coloring rules: {}.\n'
            '( 1 <= th1 < th2 <= 99, defaults: {} {} )'
        ).format(coloring_rules, *gpu_thresholds),
    )
    memory_thresholds = Device.MEMORY_UTILIZATION_THRESHOLDS
    coloring.add_argument(
        '--mem-util-thresh',
        type=int,
        nargs=2,
        choices=range(1, 100),
        metavar=('th1', 'th2'),
        help=(
            'Thresholds of GPU memory percent to determine the load intensity.\n'
            'Coloring rules: {}.\n'
            '( 1 <= th1 < th2 <= 99, defaults: {} {} )'
        ).format(coloring_rules, *memory_thresholds),
    )

    device_filtering = parser.add_argument_group('device filtering')
    device_filtering.add_argument(
        '--only',
        '-o',
        dest='only',
        type=int,
        nargs='+',
        metavar='idx',
        help='Only show the specified devices, suppress option `--only-visible`.',
    )
    device_filtering.add_argument(
        '--only-visible',
        '-ov',
        dest='only_visible',
        action='store_true',
        help='Only show devices in the `CUDA_VISIBLE_DEVICES` environment variable.',
    )

    process_filtering = parser.add_argument_group('process filtering')
    process_filtering.add_argument(
        '--compute',
        '-c',
        dest='compute',
        action='store_true',
        help="Only show GPU processes with the compute context. (type: 'C' or 'C+G')",
    )
    process_filtering.add_argument(
        '--only-compute',
        '-C',
        dest='only_compute',
        action='store_true',
        help="Only show GPU processes exactly with the compute context. (type: 'C' only)",
    )
    process_filtering.add_argument(
        '--graphics',
        '-g',
        dest='graphics',
        action='store_true',
        help="Only show GPU processes with the graphics context. (type: 'G' or 'C+G')",
    )
    process_filtering.add_argument(
        '--only-graphics',
        '-G',
        dest='only_graphics',
        action='store_true',
        help="Only show GPU processes exactly with the graphics context. (type: 'G' only)",
    )
    process_filtering.add_argument(
        '--user',
        '-u',
        dest='user',
        type=str,
        nargs='*',
        metavar='USERNAME',
        help='Only show processes of the given users (or `$USER` for no argument).',
    )
    process_filtering.add_argument(
        '--pid',
        '-p',
        dest='pid',
        type=int,
        nargs='+',
        metavar='PID',
        help='Only show processes of the given PIDs.',
    )

    args = parser.parse_args()

    if args.interval is not None and args.interval < 0.25:
        parser.error(
            f'the interval {args.interval:0.2g}s is too short, which may cause performance issues. '
            f'Expected 1/4 or higher.',
        )

    if not args.colorful:
        args.colorful = 'colorful' in NVITOP_MONITOR_MODE and 'plain' not in NVITOP_MONITOR_MODE
    if not args.light:
        args.light = 'light' in NVITOP_MONITOR_MODE and 'dark' not in NVITOP_MONITOR_MODE
    if args.user is not None and len(args.user) == 0:
        args.user.append(USERNAME)
    if args.gpu_util_thresh is None:
        try:
            gpu_util_thresh = list(
                map(int, os.getenv('NVITOP_GPU_UTILIZATION_THRESHOLDS', '').split(',')),
            )[:2]
            if (
                len(gpu_util_thresh) != 2
                or min(gpu_util_thresh) <= 0
                or max(gpu_util_thresh) >= 100
            ):
                raise ValueError
        except ValueError:
            pass
        else:
            args.gpu_util_thresh = gpu_util_thresh
    if args.mem_util_thresh is None:
        try:
            mem_util_thresh = list(
                map(int, os.getenv('NVITOP_MEMORY_UTILIZATION_THRESHOLDS', '').split(',')),
            )[:2]
            if (
                len(mem_util_thresh) != 2
                or min(mem_util_thresh) <= 0
                or max(mem_util_thresh) >= 100
            ):
                raise ValueError
        except ValueError:
            pass
        else:
            args.mem_util_thresh = mem_util_thresh

    return args


# pylint: disable-next=too-many-branches,too-many-statements,too-many-locals
def main() -> int:
    """Main function for ``nvitop`` CLI."""
    args = parse_arguments()

    if args.force_color:
        set_color(True)

    messages = []
    if args.once and hasattr(args, 'monitor'):
        messages.append('ERROR: Both `--once` and `--monitor` switches are on.')
        del args.monitor

    if not args.once and not hasattr(args, 'monitor') and TTY:
        args.monitor = None

    if hasattr(args, 'monitor') and not TTY:
        messages.append('ERROR: You must run monitor mode from a TTY terminal.')
        del args.monitor

    if hasattr(args, 'monitor') and args.monitor is None:
        mode = NVITOP_MONITOR_MODE.intersection({'auto', 'full', 'compact'})
        mode = 'auto' if len(mode) != 1 else mode.pop()
        args.monitor = mode

    if not setlocale_utf8():
        args.ascii = True

    try:
        device_count = Device.count()
    except libnvml.NVMLError_LibraryNotFound:
        return 1
    except libnvml.NVMLError as ex:
        print(
            '{} {}'.format(colored('NVML ERROR:', color='red', attrs=('bold',)), ex),
            file=sys.stderr,
        )
        return 1

    if args.gpu_util_thresh is not None:
        Device.GPU_UTILIZATION_THRESHOLDS = tuple(sorted(args.gpu_util_thresh))
    if args.mem_util_thresh is not None:
        Device.MEMORY_UTILIZATION_THRESHOLDS = tuple(sorted(args.mem_util_thresh))

    if args.only is not None:
        indices = set(args.only)
        invalid_indices = indices.difference(range(device_count))
        indices.intersection_update(range(device_count))
        if len(invalid_indices) > 1:
            messages.append(f'ERROR: Invalid device indices: {sorted(invalid_indices)}.')
        elif len(invalid_indices) == 1:
            messages.append(f'ERROR: Invalid device index: {next(iter(invalid_indices))}.')
    elif args.only_visible:
        indices = {
            index if isinstance(index, int) else index[0]
            for index in Device.parse_cuda_visible_devices()
        }
    else:
        indices = set(range(device_count))
    devices = Device.from_indices(sorted(indices))

    filters = []
    if args.compute:
        filters.append(lambda process: 'C' in process.type or 'X' in process.type)
    if args.only_compute:
        filters.append(lambda process: 'G' not in process.type and 'X' not in process.type)
    if args.graphics:
        filters.append(lambda process: 'G' in process.type or 'X' in process.type)
    if args.only_graphics:
        filters.append(lambda process: 'C' not in process.type and 'X' not in process.type)
    if args.user is not None:
        users = set(args.user)
        filters.append(lambda process: process.username in users)
    if args.pid is not None:
        pids = set(args.pid)
        filters.append(lambda process: process.pid in pids)

    ui = None
    if hasattr(args, 'monitor') and len(devices) > 0:
        try:
            with libcurses(colorful=args.colorful, light_theme=args.light) as win:
                ui = UI(
                    devices,
                    filters,
                    ascii=args.ascii,
                    mode=args.monitor,
                    interval=args.interval,
                    win=win,
                )
                ui.loop()
        except curses.error as ex:
            if ui is not None:
                raise
            messages.append(f'ERROR: Failed to initialize `curses` ({ex})')

    if ui is None:
        ui = UI(devices, filters, ascii=args.ascii)
        if not sys.stdout.isatty():
            parent = HostProcess().parent()
            if parent is not None:
                grandparent = parent.parent()
                if (
                    grandparent is not None
                    and parent.name() == 'sh'
                    and grandparent.name() == 'watch'
                ):
                    messages.append(
                        'HINT: You are running `nvitop` under `watch` command. '
                        'Please try `nvitop -m` directly.',
                    )

    ui.print()
    ui.destroy()

    if len(libnvml.UNKNOWN_FUNCTIONS) > 0:
        unknown_function_messages = [
            'ERROR: Some FunctionNotFound errors occurred while calling:'
            if len(libnvml.UNKNOWN_FUNCTIONS) > 1
            else 'ERROR: A FunctionNotFound error occurred while calling:',
        ]
        unknown_function_messages.extend(
            f'    nvmlQuery({(func.__name__ if not isinstance(func, str) else func)!r}, *args, **kwargs)'
            for func, _ in libnvml.UNKNOWN_FUNCTIONS.values()
        )
        unknown_function_messages.append(
            (
                'Please verify whether the `nvidia-ml-py` package is compatible with your NVIDIA driver version.\n'
                'You can check the release history of `nvidia-ml-py` and install the compatible version manually.\n'
                'See {} for more information.'
            ).format(
                colored('https://github.com/XuehaiPan/nvitop#installation', attrs=('underline',)),
            ),
        )

    if libnvml._pynvml_installation_corrupted:  # pylint: disable=protected-access
        message = '\n'.join(
            (
                'WARNING: The `nvidia-ml-py` package is corrupted. Please reinstall it using:',
                '',
                '    pip3 install --force-reinstall nvitop nvidia-ml-py',
                '',
                'or install `nvitop` in an isolated environment:',
                '',
                '    pip3 install --upgrade pipx',
                '    pipx install nvitop',
                '',
            ),
        )
        messages.append(message)

    if len(messages) > 0:
        for message in messages:
            if message.startswith('ERROR:'):
                message = message.replace(
                    'ERROR:',
                    colored('ERROR:', color='red', attrs=('bold',)),
                    1,
                )
            elif message.startswith('WARNING:'):
                message = message.replace(
                    'WARNING:',
                    colored('WARNING:', color='yellow', attrs=('bold',)),
                    1,
                )
            elif message.startswith('HINT:'):
                message = message.replace(
                    'HINT:',
                    colored('HINT:', color='green', attrs=('bold',)),
                    1,
                )
            print(message, file=sys.stderr)
        return 1
    return 0


if __name__ == '__main__':
    sys.exit(main())



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
#
# Copyright 2021-2023 Xuehai Pan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""The live classes for process running on the host and the GPU devices."""

# pylint: disable=too-many-lines

from __future__ import annotations

import contextlib
import datetime
import functools
import os
import threading
from abc import ABCMeta
from types import FunctionType
from typing import TYPE_CHECKING, Any, Callable, Generator, Iterable
from weakref import WeakValueDictionary

from nvitop.api import host, libnvml
from nvitop.api.utils import (
    NA,
    UINT_MAX,
    NaType,
    Snapshot,
    bytes2human,
    memoize_when_activated,
    timedelta2human,
)


if TYPE_CHECKING:
    from typing_extensions import Self  # Python 3.11+

    from nvitop.api.device import Device


__all__ = ['HostProcess', 'GpuProcess', 'command_join']


if host.POSIX:

    def add_quotes(s: str) -> str:
        """Return a shell-escaped version of the string."""
        if s == '':
            return '""'
        if '$' not in s and '\\' not in s and '\n' not in s:
            if ' ' not in s:
                return s
            if '"' not in s:
                return f'"{s}"'
        if "'" not in s and '\n' not in s:
            return f"'{s}'"
        return '"{}"'.format(
            s.replace('\\', r'\\').replace('"', r'\"').replace('$', r'\$').replace('\n', r'\n'),
        )

elif host.WINDOWS:

    def add_quotes(s: str) -> str:
        """Return a shell-escaped version of the string."""
        if s == '':
            return '""'
        if '%' not in s and '^' not in s and '\n' not in s:
            if ' ' not in s:
                return s
            if '"' not in s:
                return f'"{s}"'
        return '"{}"'.format(
            s.replace('^', '^^').replace('"', '^"').replace('%', '^%').replace('\n', r'\n'),
        )

else:

    def add_quotes(s: str) -> str:
        """Return a shell-escaped version of the string."""
        return '"{}"'.format(s.replace('\n', r'\n'))


def command_join(cmdline: list[str]) -> str:
    """Return a shell-escaped string from command line arguments."""
    if len(cmdline) == 1 and not (
        # May be modified by `setproctitle`
        os.path.isfile(cmdline[0])
        and os.path.isabs(cmdline[0])
    ):
        return cmdline[0]
    return ' '.join(map(add_quotes, cmdline))


_RAISE = object()
_USE_FALLBACK_WHEN_RAISE = threading.local()  # see also `GpuProcess.failsafe`


def auto_garbage_clean(
    fallback: Any = _RAISE,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Remove the object references in the instance cache if the method call fails (the process is gone).

    The fallback value will be used with `:meth:`GpuProcess.failsafe`` context manager, otherwise
    raises an exception when falls.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        def wrapped(self: GpuProcess, *args: Any, **kwargs: Any) -> Any:
            try:
                return func(self, *args, **kwargs)
            except host.PsutilError as ex:
                try:
                    with GpuProcess.INSTANCE_LOCK:
                        del GpuProcess.INSTANCES[(self.pid, self.device)]
                except (KeyError, AttributeError):
                    pass
                try:
                    with HostProcess.INSTANCE_LOCK:
                        del HostProcess.INSTANCES[self.pid]
                except KeyError:
                    pass
                # See also `GpuProcess.failsafe`
                if fallback is _RAISE or not getattr(_USE_FALLBACK_WHEN_RAISE, 'value', False):
                    raise ex
                if isinstance(fallback, tuple):
                    if isinstance(ex, host.AccessDenied) and fallback == ('No Such Process',):
                        return ['No Permissions']
                    return list(fallback)
                return fallback

        return wrapped

    return wrapper


class HostProcess(host.Process, metaclass=ABCMeta):
    """Represent an OS process with the given PID.

    If PID is omitted current process PID (:func:`os.getpid`) is used. The instance will be cache
    during the lifetime of the process.

    Examples:
        >>> HostProcess()  # the current process
        HostProcess(pid=12345, name='python3', status='running', started='00:55:43')

        >>> p1 = HostProcess(12345)
        >>> p2 = HostProcess(12345)
        >>> p1 is p2                 # the same instance
        True

        >>> import copy
        >>> copy.deepcopy(p1) is p1  # the same instance
        True

        >>> p = HostProcess(pid=12345)
        >>> p.cmdline()
        ['python3', '-c', 'import IPython; IPython.terminal.ipapp.launch_new_instance()']
        >>> p.command()  # the result is in shell-escaped format
        'python3 -c "import IPython; IPython.terminal.ipapp.launch_new_instance()"'

        >>> p.as_snapshot()
        HostProcessSnapshot(
            real=HostProcess(pid=12345, name='python3', status='running', started='00:55:43'),
            cmdline=['python3', '-c', 'import IPython; IPython.terminal.ipapp.launch_new_instance()'],
            command='python3 -c "import IPython; IPython.terminal.ipapp.launch_new_instance()"',
            connections=[],
            cpu_percent=0.3,
            cpu_times=pcputimes(user=2.180019456, system=0.18424464, children_user=0.0, children_system=0.0),
            create_time=1656608143.31,
            cwd='/home/panxuehai',
            environ={...},
            ...
        )
    """

    INSTANCE_LOCK: threading.RLock = threading.RLock()
    INSTANCES: WeakValueDictionary[int, HostProcess] = WeakValueDictionary()

    _pid: int
    _super_gone: bool
    _username: str | None
    _ident: tuple
    _lock: threading.RLock

    def __new__(cls, pid: int | None = None) -> Self:
        """Return the cached instance of :class:`HostProcess`."""
        if pid is None:
            pid = os.getpid()

        with cls.INSTANCE_LOCK:
            try:
                instance = cls.INSTANCES[pid]
                if instance.is_running():
                    return instance
            except KeyError:
                pass

            instance = super().__new__(cls)

            instance._super_gone = False
            instance._username = None
            host.Process._init(instance, pid, True)
            try:
                host.Process.cpu_percent(instance)
            except host.PsutilError:
                pass

            cls.INSTANCES[pid] = instance

            return instance

    # pylint: disable-next=unused-argument,super-init-not-called
    def __init__(self, pid: int | None = None) -> None:
        """Initialize the instance."""

    @property
    def _gone(self) -> bool:
        return self._super_gone

    @_gone.setter
    def _gone(self, value: bool) -> None:
        if value:
            with self.INSTANCE_LOCK:
                try:
                    del self.INSTANCES[self.pid]
                except KeyError:
                    pass
        self._super_gone = value

    def __repr__(self) -> str:
        """Return a string representation of the process."""
        return super().__repr__().replace(self.__class__.__module__ + '.', '', 1)

    def __reduce__(self) -> tuple[type[HostProcess], tuple[int]]:
        """Return state information for pickling."""
        return self.__class__, (self.pid,)

    if host.WINDOWS:

        def username(self) -> str:
            """The name of the user that owns the process.

            On Windows, the domain name will be removed if it is present.

            Raises:
                host.NoSuchProcess:
                    If the process is gone.
                host.AccessDenied:
                    If the user do not have read privilege to the process' status file.
            """
            if self._username is None:  # pylint: disable=access-member-before-definition
                self._username = (  # pylint: disable=attribute-defined-outside-init
                    super().username().split('\\')[-1]
                )
            return self._username

    else:

        def username(self) -> str:
            """The name of the user that owns the process.

            On UNIX this is calculated by using *real* process uid.

            Raises:
                host.NoSuchProcess:
                    If the process is gone.
                host.AccessDenied:
                    If the user do not have read privilege to the process' status file.
            """
            if self._username is None:  # pylint: disable=access-member-before-definition
                self._username = (  # pylint: disable=attribute-defined-outside-init
                    super().username()
                )
            return self._username

    @memoize_when_activated
    def cmdline(self) -> list[str]:
        """The command line this process has been called with.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.
        """
        cmdline = super().cmdline()
        if len(cmdline) > 1:
            cmdline = '\0'.join(cmdline).rstrip('\0').split('\0')
        return cmdline

    def command(self) -> str:
        """Return a shell-escaped string from command line arguments.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.
        """
        return command_join(self.cmdline())

    @memoize_when_activated
    def running_time(self) -> datetime.timedelta:
        """The elapsed time this process has been running in :class:`datetime.timedelta`.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.
        """
        return datetime.datetime.now() - datetime.datetime.fromtimestamp(self.create_time())

    def running_time_human(self) -> str:
        """The elapsed time this process has been running in human readable format.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.
        """
        return timedelta2human(self.running_time())

    def running_time_in_seconds(self) -> float:  # in seconds
        """The elapsed time this process has been running in seconds.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.
        """
        return self.running_time().total_seconds()

    elapsed_time = running_time
    elapsed_time_human = running_time_human
    elapsed_time_in_seconds = running_time_in_seconds

    def rss_memory(self) -> int:  # in bytes
        """The used resident set size (RSS) memory of the process in bytes.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.
        """
        return self.memory_info().rss

    def parent(self) -> HostProcess | None:
        """Return the parent process as a :class:`HostProcess` instance or :data:`None` if there is no parent.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.
        """
        parent = super().parent()
        if parent is not None:
            return HostProcess(parent.pid)
        return None

    def children(self, recursive: bool = False) -> list[HostProcess]:
        """Return the children of this process as a list of :class:`HostProcess` instances.

        If *recursive* is :data:`True` return all the descendants.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.
        """
        return [HostProcess(child.pid) for child in super().children(recursive)]

    @contextlib.contextmanager
    def oneshot(self) -> Generator[None, None, None]:
        """A utility context manager which considerably speeds up the retrieval of multiple process information at the same time.

        Internally different process info (e.g. name, ppid, uids, gids, ...) may be fetched by using
        the same routine, but only one information is returned and the others are discarded. When
        using this context manager the internal routine is executed once (in the example below on
        ``name()``) and the other info are cached.

        The cache is cleared when exiting the context manager block. The advice is to use this every
        time you retrieve more than one information about the process.

        Examples:
            >>> from nvitop import HostProcess
            >>> p = HostProcess()
            >>> with p.oneshot():
            ...     p.name()         # collect multiple info
            ...     p.cpu_times()    # return cached value
            ...     p.cpu_percent()  # return cached value
            ...     p.create_time()  # return cached value
        """  # pylint: disable=line-too-long
        with self._lock:
            if hasattr(self, '_cache'):
                yield
            else:
                with super().oneshot():
                    # pylint: disable=no-member
                    try:
                        self.cmdline.cache_activate(self)  # type: ignore[attr-defined]
                        self.running_time.cache_activate(self)  # type: ignore[attr-defined]
                        yield
                    finally:
                        self.cmdline.cache_deactivate(self)  # type: ignore[attr-defined]
                        self.running_time.cache_deactivate(self)  # type: ignore[attr-defined]

    def as_snapshot(
        self,
        attrs: Iterable[str] | None = None,
        ad_value: Any | None = None,
    ) -> Snapshot:
        """Return a onetime snapshot of the process."""
        with self.oneshot():
            attributes = self.as_dict(attrs=attrs, ad_value=ad_value)

            if attrs is None:
                for attr in ('command', 'running_time', 'running_time_human'):
                    try:
                        attributes[attr] = getattr(self, attr)()
                    except (host.AccessDenied, host.ZombieProcess):
                        attributes[attr] = ad_value

        return Snapshot(real=self, **attributes)


@HostProcess.register
class GpuProcess:  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """Represent a process with the given PID running on the given GPU device.

    The instance will be cache during the lifetime of the process.

    The same host process can use multiple GPU devices. The :class:`GpuProcess` instances
    representing the same PID on the host but different GPU devices are different.
    """

    INSTANCE_LOCK: threading.RLock = threading.RLock()
    INSTANCES: WeakValueDictionary[tuple[int, Device], GpuProcess] = WeakValueDictionary()

    _pid: int
    _host: HostProcess
    _device: Device
    _username: str | None
    _ident: tuple
    _hash: int | None

    # pylint: disable-next=too-many-arguments
    def __new__(
        cls,
        pid: int | None,
        device: Device,
        *,
        # pylint: disable=unused-argument
        gpu_memory: int | NaType | None = None,
        gpu_instance_id: int | NaType | None = None,
        compute_instance_id: int | NaType | None = None,
        type: str | NaType | None = None,  # pylint: disable=redefined-builtin
        # pylint: enable=unused-argument
    ) -> Self:
        """Return the cached instance of :class:`GpuProcess`."""
        if pid is None:
            pid = os.getpid()

        with cls.INSTANCE_LOCK:
            try:
                instance = cls.INSTANCES[(pid, device)]
                if instance.is_running():
                    return instance  # type: ignore[return-value]
            except KeyError:
                pass

            instance = super().__new__(cls)

            instance._pid = pid
            instance._host = HostProcess(pid)
            instance._ident = (*instance._host._ident, device.index)
            instance._device = device

            instance._hash = None
            instance._username = None

            cls.INSTANCES[(pid, device)] = instance

            return instance

    # pylint: disable-next=too-many-arguments
    def __init__(
        self,
        pid: int | None,  # pylint: disable=unused-argument
        device: Device,
        *,
        gpu_memory: int | NaType | None = None,
        gpu_instance_id: int | NaType | None = None,
        compute_instance_id: int | NaType | None = None,
        type: str | NaType | None = None,  # pylint: disable=redefined-builtin
    ) -> None:
        """Initialize the instance returned by :meth:`__new__()`."""
        if gpu_memory is None and not hasattr(self, '_gpu_memory'):
            gpu_memory = NA
        if gpu_memory is not None:
            self.set_gpu_memory(gpu_memory)

        if type is None and not hasattr(self, '_type'):
            type = NA
        if type is not None:
            self.type = type

        if gpu_instance_id is not None and compute_instance_id is not None:
            self._gpu_instance_id = gpu_instance_id if gpu_instance_id != UINT_MAX else NA
            self._compute_instance_id = (
                compute_instance_id if compute_instance_id != UINT_MAX else NA
            )
        elif device.is_mig_device():
            self._gpu_instance_id = device.gpu_instance_id()
            self._compute_instance_id = device.compute_instance_id()
        else:
            self._gpu_instance_id = self._compute_instance_id = NA

        for util in ('sm', 'memory', 'encoder', 'decoder'):
            if not hasattr(self, f'_gpu_{util}_utilization'):
                setattr(self, f'_gpu_{util}_utilization', NA)

    def __repr__(self) -> str:
        """Return a string representation of the GPU process."""
        return '{}(pid={}, gpu_memory={}, type={}, device={}, host={})'.format(
            self.__class__.__name__,
            self.pid,
            self.gpu_memory_human(),
            self.type,
            self.device,
            self.host,
        )

    def __eq__(self, other: object) -> bool:
        """Test equality to other object."""
        if not isinstance(other, (GpuProcess, host.Process)):
            return NotImplemented
        return self._ident == other._ident

    def __hash__(self) -> int:
        """Return a hash value of the GPU process."""
        if self._hash is None:  # pylint: disable=access-member-before-definition
            self._hash = hash(self._ident)  # pylint: disable=attribute-defined-outside-init
        return self._hash

    def __getattr__(self, name: str) -> Any | Callable[..., Any]:
        """Get a member from the instance or fallback to the host process instance if missing.

        Raises:
            AttributeError:
                If the attribute is not defined in either :class:`GpuProcess` nor :class:`HostProcess`.
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.
        """
        try:
            return super().__getattr__(name)  # type: ignore[misc]
        except AttributeError:
            if name == '_cache':
                raise
            attribute = getattr(self.host, name)
            if isinstance(attribute, FunctionType):
                attribute = auto_garbage_clean(fallback=_RAISE)(attribute)

            setattr(self, name, attribute)
            return attribute

    @property
    def pid(self) -> int:
        """The process PID."""
        return self._pid

    @property
    def host(self) -> HostProcess:
        """The process instance running on the host."""
        return self._host

    @property
    def device(self) -> Device:
        """The GPU device the process running on.

        The same host process can use multiple GPU devices. The :class:`GpuProcess` instances
        representing the same PID on the host but different GPU devices are different.
        """
        return self._device

    def gpu_instance_id(self) -> int | NaType:
        """The GPU instance ID of the MIG device, or :const:`nvitop.NA` if not applicable."""
        return self._gpu_instance_id

    def compute_instance_id(self) -> int | NaType:
        """The compute instance ID of the MIG device, or :const:`nvitop.NA` if not applicable."""
        return self._compute_instance_id

    def gpu_memory(self) -> int | NaType:  # in bytes
        """The used GPU memory in bytes, or :const:`nvitop.NA` if not applicable."""
        return self._gpu_memory

    def gpu_memory_human(self) -> str | NaType:  # in human readable
        """The used GPU memory in human readable format, or :const:`nvitop.NA` if not applicable."""
        return self._gpu_memory_human

    def gpu_memory_percent(self) -> float | NaType:  # in percentage
        """The percentage of used GPU memory by the process, or :const:`nvitop.NA` if not applicable."""
        return self._gpu_memory_percent

    def gpu_sm_utilization(self) -> int | NaType:  # in percentage
        """The utilization rate of SM (Streaming Multiprocessor), or :const:`nvitop.NA` if not applicable."""
        return self._gpu_sm_utilization

    def gpu_memory_utilization(self) -> int | NaType:  # in percentage
        """The utilization rate of GPU memory bandwidth, or :const:`nvitop.NA` if not applicable."""
        return self._gpu_memory_utilization

    def gpu_encoder_utilization(self) -> int | NaType:  # in percentage
        """The utilization rate of the encoder, or :const:`nvitop.NA` if not applicable."""
        return self._gpu_encoder_utilization

    def gpu_decoder_utilization(self) -> int | NaType:  # in percentage
        """The utilization rate of the decoder, or :const:`nvitop.NA` if not applicable."""
        return self._gpu_decoder_utilization

    def set_gpu_memory(self, value: int | NaType) -> None:
        """Set the used GPU memory in bytes."""
        # pylint: disable=attribute-defined-outside-init
        self._gpu_memory = memory_used = value
        self._gpu_memory_human = bytes2human(self.gpu_memory())
        memory_total = self.device.memory_total()
        gpu_memory_percent = NA
        if libnvml.nvmlCheckReturn(memory_used, int) and libnvml.nvmlCheckReturn(memory_total, int):
            gpu_memory_percent = round(100.0 * memory_used / memory_total, 1)  # type: ignore[assignment]
        self._gpu_memory_percent = gpu_memory_percent

    def set_gpu_utilization(
        self,
        gpu_sm_utilization: int | NaType | None = None,
        gpu_memory_utilization: int | NaType | None = None,
        gpu_encoder_utilization: int | NaType | None = None,
        gpu_decoder_utilization: int | NaType | None = None,
    ) -> None:
        """Set the GPU utilization rates."""
        # pylint: disable=attribute-defined-outside-init
        if gpu_sm_utilization is not None:
            self._gpu_sm_utilization = gpu_sm_utilization
        if gpu_memory_utilization is not None:
            self._gpu_memory_utilization = gpu_memory_utilization
        if gpu_encoder_utilization is not None:
            self._gpu_encoder_utilization = gpu_encoder_utilization
        if gpu_decoder_utilization is not None:
            self._gpu_decoder_utilization = gpu_decoder_utilization

    def update_gpu_status(self) -> int | NaType:
        """Update the GPU consumption status from a new NVML query."""
        self.set_gpu_memory(NA)
        self.set_gpu_utilization(NA, NA, NA, NA)
        processes = self.device.processes()
        process = processes.get(self.pid, self)
        if process is not self:
            # The current process is gone and the instance has been removed from the cache.
            # Update GPU status from the new instance.
            self.set_gpu_memory(process.gpu_memory())
            self.set_gpu_utilization(
                process.gpu_sm_utilization(),
                process.gpu_memory_utilization(),
                process.gpu_encoder_utilization(),
                process.gpu_decoder_utilization(),
            )
        return self.gpu_memory()

    @property
    def type(self) -> str | NaType:
        """The type of the GPU context.

        The type is one of the following:
            - :data:`'C'`: compute context
            - :data:`'G'`: graphics context
            - :data:`'C+G'`: both compute context and graphics context
            - :data:`'N/A'`: not applicable
        """
        return self._type

    @type.setter
    def type(self, value: str | NaType) -> None:
        if 'C' in value and 'G' in value:
            self._type = 'C+G'
        elif 'C' in value:
            self._type = 'C'
        elif 'G' in value:
            self._type = 'G'
        else:
            self._type = NA

    @auto_garbage_clean(fallback=False)
    def is_running(self) -> bool:
        """Return whether this process is running."""
        return self.host.is_running()

    @auto_garbage_clean(fallback='terminated')
    def status(self) -> str:
        """The process current status.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        return self.host.status()

    @auto_garbage_clean(fallback=NA)
    def create_time(self) -> float | NaType:
        """The process creation time as a floating point number expressed in seconds since the epoch.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        return self.host.create_time()

    @auto_garbage_clean(fallback=NA)
    def running_time(self) -> datetime.timedelta | NaType:
        """The elapsed time this process has been running in :class:`datetime.timedelta`.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        return self.host.running_time()

    def running_time_human(self) -> str | NaType:
        """The elapsed time this process has been running in human readable format.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        return timedelta2human(self.running_time())

    def running_time_in_seconds(self) -> float | NaType:
        """The elapsed time this process has been running in seconds.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        running_time = self.running_time()
        if running_time is NA:
            return NA
        return running_time.total_seconds()

    elapsed_time = running_time
    elapsed_time_human = running_time_human
    elapsed_time_in_seconds = running_time_in_seconds

    @auto_garbage_clean(fallback=NA)
    def username(self) -> str | NaType:
        """The name of the user that owns the process.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        if self._username is None:  # pylint: disable=access-member-before-definition
            self._username = self.host.username()  # pylint: disable=attribute-defined-outside-init
        return self._username

    @auto_garbage_clean(fallback=NA)
    def name(self) -> str | NaType:
        """The process name.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        return self.host.name()

    @auto_garbage_clean(fallback=NA)
    def cpu_percent(self) -> float | NaType:  # in percentage
        """Return a float representing the current process CPU utilization as a percentage.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        return self.host.cpu_percent()

    @auto_garbage_clean(fallback=NA)
    def memory_percent(self) -> float | NaType:  # in percentage
        """Compare process RSS memory to total physical system memory and calculate process memory utilization as a percentage.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """  # pylint: disable=line-too-long
        return self.host.memory_percent()

    host_memory_percent = memory_percent  # in percentage

    @auto_garbage_clean(fallback=NA)
    def host_memory(self) -> int | NaType:  # in bytes
        """The used resident set size (RSS) memory of the process in bytes.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        return self.host.rss_memory()

    def host_memory_human(self) -> str | NaType:
        """The used resident set size (RSS) memory of the process in human readable format.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        return bytes2human(self.host_memory())

    rss_memory = host_memory  # in bytes

    # For `AccessDenied` error the fallback value is `['No Permissions']`
    @auto_garbage_clean(fallback=('No Such Process',))
    def cmdline(self) -> list[str]:
        """The command line this process has been called with.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        cmdline = self.host.cmdline()
        if len(cmdline) == 0 and not self._gone:
            cmdline = ['Zombie Process']
        return cmdline

    def command(self) -> str:
        """Return a shell-escaped string from command line arguments.

        Raises:
            host.NoSuchProcess:
                If the process is gone.
            host.AccessDenied:
                If the user do not have read privilege to the process' status file.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        return command_join(self.cmdline())

    @auto_garbage_clean(fallback=_RAISE)
    def host_snapshot(self) -> Snapshot:
        """Return a onetime snapshot of the host process."""
        with self.host.oneshot():
            return Snapshot(
                real=self.host,
                is_running=self.is_running(),
                status=self.status(),
                username=self.username(),
                name=self.name(),
                cmdline=self.cmdline(),
                command=self.command(),
                cpu_percent=self.cpu_percent(),
                memory_percent=self.memory_percent(),
                host_memory=self.host_memory(),
                host_memory_human=self.host_memory_human(),
                running_time=self.running_time(),
                running_time_human=self.running_time_human(),
                running_time_in_seconds=self.running_time_in_seconds(),
            )

    @auto_garbage_clean(fallback=_RAISE)
    def as_snapshot(
        self,
        *,
        host_process_snapshot_cache: dict[int, Snapshot] | None = None,
    ) -> Snapshot:
        """Return a onetime snapshot of the process on the GPU device.

        Note:
            To return the fallback value rather than raise an exception, please use the context
            manager :meth:`GpuProcess.failsafe`. Also, consider using the batched version to take
            snapshots with :meth:`GpuProcess.take_snapshots`, which caches the results and reduces
            redundant queries. See also :meth:`take_snapshots` and :meth:`failsafe`.
        """
        host_process_snapshot_cache = host_process_snapshot_cache or {}
        try:
            host_snapshot = host_process_snapshot_cache[self.pid]
        except KeyError:
            host_snapshot = host_process_snapshot_cache[self.pid] = self.host_snapshot()

        return Snapshot(
            real=self,
            pid=self.pid,
            # host
            host=host_snapshot,
            is_running=host_snapshot.is_running,
            status=host_snapshot.status,
            username=host_snapshot.username,
            name=host_snapshot.name,
            cmdline=host_snapshot.cmdline,
            command=host_snapshot.command,
            cpu_percent=host_snapshot.cpu_percent,
            memory_percent=host_snapshot.memory_percent,
            host_memory=host_snapshot.host_memory,
            host_memory_human=host_snapshot.host_memory_human,
            running_time=host_snapshot.running_time,
            running_time_human=host_snapshot.running_time_human,
            running_time_in_seconds=host_snapshot.running_time_in_seconds,
            # device
            device=self.device,
            type=self.type,
            gpu_instance_id=self.gpu_instance_id(),
            compute_instance_id=self.compute_instance_id(),
            gpu_memory=self.gpu_memory(),
            gpu_memory_human=self.gpu_memory_human(),
            gpu_memory_percent=self.gpu_memory_percent(),
            gpu_sm_utilization=self.gpu_sm_utilization(),
            gpu_memory_utilization=self.gpu_memory_utilization(),
            gpu_encoder_utilization=self.gpu_encoder_utilization(),
            gpu_decoder_utilization=self.gpu_decoder_utilization(),
        )

    @classmethod
    def take_snapshots(  # batched version of `as_snapshot`
        cls,
        gpu_processes: Iterable[GpuProcess],
        *,
        failsafe: bool = False,
    ) -> list[Snapshot]:
        """Take snapshots for a list of :class:`GpuProcess` instances.

        If *failsafe* is :data:`True`, then if any method fails, the fallback value in
        :func:`auto_garbage_clean` will be used.
        """
        cache: dict[int, Snapshot] = {}
        context: Callable[[], contextlib.AbstractContextManager[None]] = (
            cls.failsafe if failsafe else contextlib.nullcontext  # type: ignore[assignment]
        )
        with context():
            return [
                process.as_snapshot(host_process_snapshot_cache=cache) for process in gpu_processes
            ]

    @classmethod
    @contextlib.contextmanager
    def failsafe(cls) -> Generator[None, None, None]:
        """A context manager that enables fallback values for methods that fail.

        Examples:
            >>> p = GpuProcess(pid=10000, device=Device(0))  # process does not exist
            >>> p
            GpuProcess(pid=10000, gpu_memory=N/A, type=N/A, device=PhysicalDevice(index=0, name="NVIDIA GeForce RTX 3070", total_memory=8192MiB), host=HostProcess(pid=10000, status='terminated'))
            >>> p.cpu_percent()
            Traceback (most recent call last):
                ...
            NoSuchProcess: process no longer exists (pid=10000)

            >>> # Failsafe to the fallback value instead of raising exceptions
            ... with GpuProcess.failsafe():
            ...     print('fallback:              {!r}'.format(p.cpu_percent()))
            ...     print('fallback (float cast): {!r}'.format(float(p.cpu_percent())))  # `nvitop.NA` can be cast to float or int
            ...     print('fallback (int cast):   {!r}'.format(int(p.cpu_percent())))    # `nvitop.NA` can be cast to float or int
            fallback:              'N/A'
            fallback (float cast): nan
            fallback (int cast):   0
        """  # pylint: disable=line-too-long
        global _USE_FALLBACK_WHEN_RAISE  # pylint: disable=global-statement,global-variable-not-assigned

        prev_value = getattr(_USE_FALLBACK_WHEN_RAISE, 'value', False)
        try:
            _USE_FALLBACK_WHEN_RAISE.value = True
            yield
        finally:
            _USE_FALLBACK_WHEN_RAISE.value = prev_value
